#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.


from typing import Any, Callable, Dict

from classy_vision.generic.distributed_util import get_rank, get_world_size
from torch.utils.data.distributed import DistributedSampler

from . import register_dataset
from .classy_video_dataset import ClassyVideoDataset
from .core import RandomVideoDataset
from .transforms.util_video import build_video_field_transform_default


@register_dataset("synthetic_video")
class SyntheticVideoDataset(ClassyVideoDataset):
    """Classy Dataset which produces random synthetic video clips.

    Useful for testing since the dataset is much faster to initialize and fetch samples
    from, compared to real world datasets.

    Note: Unlike :class:`SyntheticImageDataset`, this dataset generates targets
        randomly, independent of the video clips.
    """

    def __init__(
        self,
        num_classes: int,
        split: str,
        batchsize_per_replica: int,
        shuffle: bool,
        transform: Callable,
        num_samples: int,
        frames_per_clip: int,
        video_width: int,
        video_height: int,
        audio_samples: int,
        clips_per_video: int,
    ):
        """The constructor of SyntheticVideoDataset.

        Args:
            num_classes: Number of classes in the generated targets.
            split: Split of dataset to use
            batchsize_per_replica: batch size per model replica
            shuffle: Whether we should shuffle between epochs
            transform: Transform to be applied to each sample
            num_samples: Number of samples to return
            frames_per_clip: Number of frames in a video clip
            video_width: Width of the video clip
            video_height: Height of the video clip
            audio_samples: Audio sample rate
            clips_per_video: Number of clips per video
        """
        dataset = RandomVideoDataset(
            num_classes,
            split,
            num_samples,
            frames_per_clip,
            video_width,
            video_height,
            audio_samples,
            clips_per_video,
        )
        super().__init__(
            dataset,
            split,
            batchsize_per_replica,
            shuffle,
            transform,
            num_samples,
            clips_per_video,
        )

    @classmethod
    def from_config(cls, config: Dict[str, Any]) -> "SyntheticVideoDataset":
        """Instantiates a SyntheticVideoDataset from a configuration.

        Args:
            config: A configuration for a SyntheticVideoDataset.
                See :func:`__init__` for parameters expected in the config.

        Returns:
            A SyntheticVideoDataset instance.
        """
        split = config["split"]
        num_classes = config["num_classes"]
        (
            transform_config,
            batchsize_per_replica,
            shuffle,
            num_samples,
            frames_per_clip,
            video_width,
            video_height,
            video_min_dimension,
            audio_samples,
            step_between_clips,
            frame_rate,
            clips_per_video,
        ) = cls.parse_config(config)

        transform = build_video_field_transform_default(transform_config, split)
        return cls(
            num_classes,
            split,
            batchsize_per_replica,
            shuffle,
            transform,
            num_samples,
            frames_per_clip,
            video_width,
            video_height,
            audio_samples,
            clips_per_video,
        )

    @property
    def video_clips(self):
        raise NotImplementedError()

    def _get_sampler(self, epoch):
        world_size = get_world_size()
        rank = get_rank()
        sampler = DistributedSampler(
            self, num_replicas=world_size, rank=rank, shuffle=self.shuffle
        )
        sampler.set_epoch(epoch)
        return sampler



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from __future__ import absolute_import, division, print_function, unicode_literals

import pathlib
import sys
import unittest

import classy_vision
from classy_vision.dataset.transforms import GenericImageTransform
from classy_vision.optim import SGD
from classy_vision.optim.param_scheduler import LinearParamScheduler
from classy_vision.tasks import ClassificationTask
from classy_vision.trainer import LocalTrainer
from torchvision import transforms


# import the classes from the synthetic template
path = pathlib.Path(classy_vision.__file__).resolve().parent
synthetic_template_path = path / "templates" / "synthetic"
sys.path.append(str(synthetic_template_path))

from datasets.my_dataset import MyDataset  # isort:skip
from losses.my_loss import MyLoss  # isort:skip
from models.my_model import MyModel  # isort:skip


# WARNING: The goal of this test is to use our public API as advertised in our
# tutorials and make sure everything trains successfully. If you break this
# test, make sure you also update our tutorials.


class APITest(unittest.TestCase):
    def test_one(self):
        train_dataset = MyDataset(
            batchsize_per_replica=32,
            shuffle=False,
            transform=GenericImageTransform(
                transform=transforms.Compose(
                    [
                        transforms.RandomResizedCrop(224),
                        transforms.RandomHorizontalFlip(),
                        transforms.ToTensor(),
                        transforms.Normalize(
                            mean=[0.485, 0.456, 0.406], std=[0.229, 0.224, 0.225]
                        ),
                    ]
                )
            ),
            num_samples=100,
            crop_size=224,
            class_ratio=0.5,
            seed=0,
        )

        test_dataset = MyDataset(
            batchsize_per_replica=32,
            shuffle=False,
            transform=GenericImageTransform(
                transform=transforms.Compose(
                    [
                        transforms.Resize(256),
                        transforms.CenterCrop(224),
                        transforms.ToTensor(),
                        transforms.Normalize(
                            mean=[0.485, 0.456, 0.406], std=[0.229, 0.224, 0.225]
                        ),
                    ]
                )
            ),
            num_samples=100,
            crop_size=224,
            class_ratio=0.5,
            seed=0,
        )

        model = MyModel()
        loss = MyLoss()

        optimizer = SGD(momentum=0.9, weight_decay=1e-4, nesterov=True)
        optimizer.set_param_schedulers(
            {"lr": LinearParamScheduler(start_lr=0.01, end_lr=0.009)}
        )

        task = (
            ClassificationTask()
            .set_model(model)
            .set_dataset(train_dataset, "train")
            .set_dataset(test_dataset, "test")
            .set_loss(loss)
            .set_optimizer(optimizer)
            .set_num_epochs(1)
        )

        trainer = LocalTrainer()
        trainer.train(task)



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from typing import Any, Dict

from classy_vision.heads import ClassyHead, register_head


@register_head("identity")
class IdentityHead(ClassyHead):
    """This head returns the input without changing it. This can
    be attached to a model, if the output of the model is the
    desired result.
    """

    def forward(self, x):
        return x

    @classmethod
    def from_config(cls, config: Dict[str, Any]) -> "IdentityHead":
        """Instantiates a IdentityHead from a configuration.

        Args:
            config: A configuration for a IdentityHead.
                See :func:`__init__` for parameters expected in the config.

        Returns:
            A IdentityHead instance.
        """
        return cls(config["unique_id"])



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import io
import logging
import os
from typing import Any, Optional

import numpy
import torch
import torchelastic
import torchelastic.distributed as dist
from classy_vision.generic.distributed_util import set_cpu_device, set_cuda_device_index
from classy_vision.generic.util import get_checkpoint_dict
from classy_vision.hooks import ClassyHookFunctions
from classy_vision.tasks import ClassyTask
from classy_vision.trainer import ClassyTrainer
from torchelastic.worker_stats import WorkerStats


log = logging.getLogger(__name__)


class ElasticTrainer(ClassyTrainer):
    def __init__(
        self,
        use_gpu,
        num_dataloader_workers,
        elastic_coordinator,
        input_args,
        local_rank,
        dataloader_mp_context=None,
    ):
        super().__init__(
            use_gpu=use_gpu,
            num_dataloader_workers=num_dataloader_workers,
            dataloader_mp_context=dataloader_mp_context,
        )
        pid = os.getpid()
        if use_gpu:
            set_cuda_device_index(local_rank)
            device_idx = torch.cuda.current_device()
            log.info(f"initialized worker {local_rank} (pid={pid}, gpu={device_idx})")
            device_properties = torch.cuda.get_device_properties(device_idx)
            log.info(f"gpu device properties: {device_properties}")
        else:
            # cpu
            set_cpu_device()
            log.info(f"initialized worker {local_rank} (pid={pid}, cpu)")

        self.elastic_coordinator = elastic_coordinator
        self.input_args = input_args

    def train(self, task):
        """
        Runs training phases, phases are generated from the config.
        """

        assert isinstance(task, ClassyTask)
        pin_memory = self.use_gpu and torch.cuda.device_count() > 1

        task.prepare(
            num_dataloader_workers=self.num_dataloader_workers,
            pin_memory=pin_memory,
            use_gpu=self.use_gpu,
            dataloader_mp_context=self.dataloader_mp_context,
        )
        state = self._ClassyElasticState(task, self.input_args)

        local_variables = {}

        state.advance_to_next_phase = True

        def elastic_train_step(orig_state):
            if state.run_start_hooks:
                # need this to ensure we don't run the on_start hooks every time
                # a trainer starts
                state.task.on_start(local_variables)
                state.run_start_hooks = False
                return state, self._ClassyWorkerStats(None)

            return self._run_step(orig_state, local_variables, self.use_gpu)

        torchelastic.train(self.elastic_coordinator, elastic_train_step, state)

        task.on_end(local_variables)

    def _run_step(self, state, local_variables, use_gpu):
        # Check for training complete but only terminate when the last phase is done
        if state.task.done_training() and state.advance_to_next_phase:
            raise StopIteration

        if state.advance_to_next_phase:
            self.elastic_coordinator.barrier()
            self.elastic_coordinator._log_event("on_phase_start")
            state.task.on_phase_start(local_variables)

            state.advance_to_next_phase = False

        # Process one train step
        try:
            if state.skip_current_phase:
                state.advance_to_next_phase = True
                state.skip_current_phase = False  # Reset flag
            else:
                state.task.step(use_gpu, local_variables)
        except StopIteration:
            state.advance_to_next_phase = True

        if state.advance_to_next_phase:
            self.elastic_coordinator.barrier()
            state.task.on_phase_end(local_variables)

        progress_rate = None  # using None to signal 'unknown'
        perf_stats = local_variables.get("perf_stats", None)
        if perf_stats is not None:
            batch_time = perf_stats._cuda_stats["train_step_total"].smoothed_value
            if batch_time is not None and batch_time > 0.0:
                # rate = number of mini-batches per second
                progress_rate = 1.0 / batch_time

        progress_stats = self._ClassyWorkerStats(progress_rate)
        return state, progress_stats

    class _ClassyWorkerStats(WorkerStats):
        """
        ClassyVision-specific implementation of WorkerStats,
        which is used by torchelastic train_loop
        to detect (and correct stragglers), or other progress-impeding issues.
        """

        def __init__(self, progress_rate):
            self.progress_rate = progress_rate

        def get_progress_rate(self) -> Optional[float]:
            return self.progress_rate

    class _ClassyElasticState(torchelastic.State):
        """
        Rollback is disabled on this state since currently, data loaders are
        too expensive to snapshot on every train_step
        """

        def __init__(self, task: ClassyTask, input_args: Any):
            # WARNING: Make sure to add any members here to self.save() and self.load()
            self.task = task
            self.input_args = input_args if input_args else {}
            self.advance_to_next_phase = True
            self.skip_current_phase = False
            self.snapshot = None
            # run_start_hooks is used to determine if the on_start hooks should be run,
            # which should happen at the beginning of training. After that, this is set
            # to False.
            self.run_start_hooks = True

        def broadcast_state(self, rank, src_rank):
            data = None
            if rank == src_rank:
                save_stream = io.BytesIO()
                self.save(save_stream)
                # Note: save_stream.getbuffer() will return a memoryview, which
                # cannot be convert to a tensor, need convert it to np array first
                data = numpy.asarray(save_stream.getbuffer())
            data = dist.broadcast_binary(data, src_rank)
            load_stream = io.BytesIO(data)
            self.load(load_stream)

        def sync(self, world_size, rank):
            self._recreate_ddp_model()

            # Figure out which trainer has the most up-to-date data, and
            # use that trainer to broadcast task to all others.
            src_rank = self._compute_most_tenured_rank(rank)
            self.broadcast_state(rank, src_rank)

            # Current on-box data loaders don't support recovery in the middle of
            # a phase and since we don't rollback the model, re-training is
            # worse than losing data so we're skipping rest of the phase.
            #
            # Also we can't just set advance_to_next_phase to True here as it
            # will cause on_phase_end() hooks to not run.
            # We also only skip the current phase if this isn't the first time
            # calling sync from the PET train_loop. We'll need to reconsider this
            # if the PET train_loop changes. advance_to_next_phase is already
            # synced from rest of the trainers at this point.
            if not self.advance_to_next_phase:
                self.skip_current_phase = True

            logging.warning(
                "RANK {}: now we all have {} updates and latest task".format(
                    rank, self.task.num_updates
                )
            )

            # Re-build dataloader, dataset, and iterator anytime membership
            # changes. When world_size potentially changes (e.g. re-rendezvous), we
            # need to re-create both the dataset and dataloader objects because we
            # create a ShardDataset based on the world size at the time of
            # construction.
            # TODO (T55691442): Figure out how to solve re-sharding without
            # rebuilding the datasets. sync() only works correctly without elasticity
            # currently.
            for phase_type in self.task.datasets.keys():
                self.task._recreate_data_loader_from_dataset(phase_type)
            self.task.create_data_iterator()
            # Set up pytorch module in train vs eval mode, update optimizer.
            self.task._set_model_train_mode()

        def should_save_checkpoint(self, rank):
            # should_save_checkpoint need to return same value for all trainers
            # we take checkpoint when a phase completed
            # TODO add test coverage for this

            # currently for typical imagenet resnet model checkpointing take 15 seconds
            # consider the cost it is not very necessary to do checkpoint for test phase
            return self.task.train and self.advance_to_next_phase

        def capture_snapshot(self):
            # Save snapshot at phase boundary. This will make sure no data-loss
            # when a failure happens. We will support fine-grade recovery once
            # fault tolerate data loader is ready
            if self.task.train and self.advance_to_next_phase:
                stream = io.BytesIO()
                self.save(stream)
                # save snapshot and return it every train step this make sure
                # we always has a good state to recover.
                self.snapshot = stream.getbuffer()
                logging.info(
                    "Take snapshot at updates {}".format(self.task.num_updates)
                )
            return self.snapshot

        def apply_snapshot(self, capture_snapshot) -> None:
            with io.BytesIO(capture_snapshot) as stream:
                self.load(stream)
                logging.info(
                    "Snapshot applied, now we are at updates {}".format(
                        self.task.num_updates
                    )
                )

        def save(self, stream):
            checkpoint_state = get_checkpoint_dict(self.task, self.input_args)
            checkpoint_state["advance_to_next_phase"] = self.advance_to_next_phase
            checkpoint_state["skip_current_phase"] = self.skip_current_phase
            checkpoint_state["run_start_hooks"] = self.run_start_hooks
            torch.save(checkpoint_state, stream)

        def load(self, stream):
            checkpoint_state = torch.load(stream, map_location=torch.device("cpu"))
            state = checkpoint_state["classy_state_dict"]
            self.task.set_classy_state(state)
            if "advance_to_next_phase" in checkpoint_state:
                self.advance_to_next_phase = checkpoint_state["advance_to_next_phase"]
            if "skip_current_phase" in checkpoint_state:
                self.skip_current_phase = checkpoint_state["skip_current_phase"]
            if "run_start_hooks" in checkpoint_state:
                self.run_start_hooks = checkpoint_state["run_start_hooks"]

        def _recreate_ddp_model(self):
            # Delete & re-create the DDP module wrapper. This is required because
            # each instance of DDP is tied to a specific process group, and
            # any time the set of workers in PET changes, we create a new
            # process group, so the old DDP wrapper is invalid.
            # TODO: does calling del here invoke C++ destructor if it's the last
            # reference? Or is assigning None sufficient?
            del self.task.distributed_model
            self.task.distributed_model = None
            self.task.init_distributed_data_parallel_model()

        def _compute_most_tenured_rank(self, rank):
            logging.warning(
                "RANK {}: syncing, I have {} updates".format(
                    rank, self.task.num_updates
                )
            )
            # Propagate state to new trainer processes.
            # First, figure out which process has a copy of the most recent
            # state by getting a copy of everybody's iteration counter.
            max_rank, max_num_updates = dist.all_gather_return_max_long(
                self.task.num_updates
            )

            logging.warning(
                "RANK {}: rank {} has the most updates {}".format(
                    rank, max_rank, max_num_updates
                )
            )

            return max_rank



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import unittest
from test.generic.optim_test_util import TestOptimizer

from classy_vision.optim.adam import Adam


class TestAdamOptimizer(TestOptimizer, unittest.TestCase):
    def _check_momentum_buffer(self):
        return False

    def _get_config(self):
        return {
            "name": "adam",
            "num_epochs": 90,
            "lr": 0.1,
            "betas": (0.9, 0.99),
            "eps": 1e-8,
            "weight_decay": 0.0001,
            "amsgrad": False,
        }

    def _instance_to_test(self):
        return Adam



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import argparse
import collections
import contextlib
import json
import logging
import math
import os
import sys
import traceback
from typing import Dict, Optional

import numpy as np
import torch
import torch.nn as nn
from fvcore.common.file_io import PathManager
from torch._six import container_abcs


# constants:
CHECKPOINT_FILE = "checkpoint.torch"
CPU_DEVICE = torch.device("cpu")


def is_pos_int(number):
    """
    Returns True if a number is a positive integer.
    """
    return type(number) == int and number >= 0


def is_pos_float(number):
    """
    Returns True if a number is a positive float.
    """
    return type(number) == float and number >= 0.0


def is_pos_int_list(l):
    """
    Returns True if a list contains positive integers
    """
    return type(l) == list and all(is_pos_int(n) for n in l)


def is_long_tensor(tensor):
    """
    Returns True if a tensor is a long tensor.
    """
    if torch.is_tensor(tensor):
        return tensor.type().endswith("LongTensor")
    else:
        return False


def is_float_tensor(tensor):
    """
    Returns True if a tensor is a float tensor.
    """
    if torch.is_tensor(tensor):
        return tensor.type().endswith("FloatTensor")
    else:
        return False


def is_double_tensor(tensor):
    """
    Returns True if a tensor is a double tensor.
    """
    if torch.is_tensor(tensor):
        return tensor.type().endswith("DoubleTensor")
    else:
        return False


def is_leaf(module):
    """
    Returns True if module is leaf in the graph.
    """
    assert isinstance(module, nn.Module), "module should be nn.Module"
    return len([c for c in module.children()]) == 0 or hasattr(module, "_mask")


def is_on_gpu(model):
    """
    Returns True if all parameters of a model live on the GPU.
    """
    assert isinstance(model, torch.nn.Module)
    on_gpu = True
    has_params = False
    for param in model.parameters():
        has_params = True
        if not param.data.is_cuda:
            on_gpu = False
    return has_params and on_gpu


def is_not_none(sample):
    """
    Returns True if sample is not None and constituents are not none.
    """
    if sample is None:
        return False

    if isinstance(sample, (list, tuple)):
        if any(s is None for s in sample):
            return False

    if isinstance(sample, dict):
        if any(s is None for s in sample.values()):
            return False
    return True


def copy_model_to_gpu(model, loss=None):
    """
    Copies a model and (optional) loss to GPU and enables cudnn benchmarking.
    For multiple gpus training, the model in DistributedDataParallel for
    distributed training.
    """
    if not torch.backends.cudnn.deterministic:
        torch.backends.cudnn.benchmark = True
    model = model.cuda()

    if loss is not None:
        loss = loss.cuda()
        return model, loss
    else:
        return model


def copy_upvalue(value, upvalue):
    """
    Iteratively copies a particular value into an upvalue dict.
    """
    assert type(value) == type(upvalue), "type of value and upvalue must match"
    if type(value) == dict:
        upvalue.clear()
        for key, val in value.items():
            upvalue[key] = val
    elif type(value) == list:
        del upvalue[:]
        for _, val in value:
            upvalue.append(val)
    else:
        raise BaseException("unsupported upvalue type")


def recursive_copy_to_gpu(value, non_blocking=True, max_depth=3, curr_depth=0):
    """
    Recursively searches lists, tuples, dicts and copies to GPU if possible.
    Note:  These are all copies, so if there are two objects that reference
    the same object, then after this call, there will be two different objects
    referenced on the GPU.
    """
    if curr_depth >= max_depth:
        raise ValueError("Depth of value object is too deep")

    try:
        return value.cuda(non_blocking=non_blocking)
    except AttributeError:
        if isinstance(value, container_abcs.Sequence):
            gpu_val = []
            for val in value:
                gpu_val.append(
                    recursive_copy_to_gpu(
                        val,
                        non_blocking=non_blocking,
                        max_depth=max_depth,
                        curr_depth=curr_depth + 1,
                    )
                )

            return gpu_val if isinstance(value, list) else tuple(gpu_val)
        elif isinstance(value, container_abcs.Mapping):
            gpu_val = {}
            for key, val in value.items():
                gpu_val[key] = recursive_copy_to_gpu(
                    val,
                    non_blocking=non_blocking,
                    max_depth=max_depth,
                    curr_depth=curr_depth + 1,
                )

            return gpu_val

        raise AttributeError("Value must have .cuda attr or be a Seq / Map iterable")


@contextlib.contextmanager
def numpy_seed(seed, *addl_seeds):
    """Context manager which seeds the NumPy PRNG with the specified seed and
    restores the state afterward"""
    if seed is None:
        yield
        return
    if len(addl_seeds) > 0:
        seed = int(hash((seed, *addl_seeds)) % 1e6)
    state = np.random.get_state()
    np.random.seed(seed)
    try:
        yield
    finally:
        np.random.set_state(state)


def accuracy(output, target, topk=(1,)):
    """
    Computes the top-k accuracy in a multi-class problem from an NxK output
    matrix and a corresponding Nx1 target matrix with values in 0, ..., K-1.
    """

    # assertions:
    assert torch.is_tensor(output)
    assert torch.is_tensor(target)
    assert output.size(0) == target.size(0)
    if type(topk) == int:
        topk = (topk,)
    assert type(topk) == tuple or type(topk) == list
    assert all(is_pos_int(k) for k in topk)
    maxk = max(topk)
    assert maxk < output.size(1)

    # determine whether predictions are correct:
    _, pred = output.topk(maxk, 1, True, True)
    pred = pred.t()
    correct = pred.eq(target.view(1, -1).expand_as(pred))

    # compute accuracies:
    res = []
    for k in topk:
        correct_k = correct[:k].view(-1).float().sum(0, keepdim=True)
        res.append(correct_k.mul_(100.0 / target.size(0)).item())
    return res


def binary_accuracy(output, target, threshold=0.0):
    """
    Computes the accuracy based on a real-valued output tensor and a corresponding
    target matrix of the same size with valuesin [0, 1].

    An optional threshold for positive classification can be specified (default = 0).
    """

    # assertions:
    assert torch.is_tensor(output)
    assert torch.is_tensor(target)
    assert output.size() == target.size()

    # compute accuracy:
    correct = output.ge(threshold).type_as(target).eq_(target).long()
    return correct.sum().item() * 100.0 / float(target.nelement())


def create_class_histograms(pred_prob, target, num_bins):
    """
    Creates two histograms to contain total positive or true samples
    for a given class at a given score value and total samples at a
    given predicted probability.  Then for a given threshold we can compute
    precision = true_samples_above_threshold /
    total_samples_above_threshold and recall =
    true_samples_above_threshold / true_samples. Output is expected to
    be prediction probabilities.

    The pred_prob vector is num_classes x batchsize_per_replica, the target vector is
    num_classes x 1, should contain integers corresponding to each class.
    Output is two tensors of same size: num_bins x num_classes
    where the bins are equispaced from 0 to 1 across the score.
    The class_hist[:, c] contains true example histogram for class c
    prediction probabilities while total_hist[:, c] contains histogram
    for all samples for class c prediction probabilities.
    """

    # assertions:
    assert torch.is_tensor(pred_prob)
    assert torch.is_tensor(target)
    assert pred_prob.size()[0] == target.size()[0], "%s pred_prob, %s target" % (
        str(pred_prob.size()),
        str(target.size()),
    )
    assert (
        pred_prob.lt(0.0).sum() == 0 and pred_prob.gt(1.0).sum() == 0
    ), "Prediction probability must be between 0 and 1"

    num_classes = pred_prob.size()[1]
    class_hist = torch.zeros([num_bins, num_classes], dtype=torch.int64)
    total_hist = torch.zeros([num_bins, num_classes], dtype=torch.int64)
    for c in range(num_classes):
        total_hist[:, c] = torch.histc(
            pred_prob[:, c], bins=num_bins, min=0.0, max=1.0
        ).long()
        class_hist[:, c] = torch.histc(
            pred_prob[:, c][(target == c).nonzero().squeeze(1)],
            bins=num_bins,
            min=0.0,
            max=1.0,
        ).long()

    return class_hist, total_hist


def _find_last_larger_than(target, val_array):
    """
    Takes an array and finds the last value larger than the target value.
    Returns the index of that value, returns -1 if none exists in array.
    """
    ind = -1
    for j in range(len(val_array), 0, -1):
        if val_array[j - 1] > target:
            ind = j - 1
            break
    return ind


def calc_ap(prec, recall):
    """
    Computes average precision from precision recall curves. Curves
    are expected to be same size and 1D.
    """
    assert prec.size() == recall.size(), "Precision and recall curves must be same size"
    ap = 0.0
    if len(prec) == 0:
        return ap
    prev_r = 0.0
    prev_p = prec[0]
    for p, r in zip(prec, recall):
        ap += (r - prev_r) * (p + prev_p) / 2.0
        prev_r = r
        prev_p = p

    return ap


def compute_pr_curves(class_hist, total_hist):
    """
    Computes precision recall curves from the true sample / total
    sample histogram tensors. The histogram tensors are num_bins x num_classes
    and each column represents a histogram over
    prediction_probabilities.

    The two tensors should have the same dimensions.
    The two tensors should have nonnegative integer values.

    Returns map of precision / recall values from highest precision to lowest
    and the calculated AUPRC (i.e. the average precision).
    """
    assert torch.is_tensor(class_hist) and torch.is_tensor(
        total_hist
    ), "Both arguments must be tensors"
    assert (
        class_hist.dtype == torch.int64 and total_hist.dtype == torch.int64
    ), "Both arguments must contain int64 values"
    assert (
        len(class_hist.size()) == 2 and len(total_hist.size()) == 2
    ), "Both arguments must have 2 dimensions, (score_bin, class)"
    assert (
        class_hist.size() == total_hist.size()
    ), """
        For compute_pr_curve, arguments must be  of same size.
        class_hist.size(): %s
        total_hist.size(): %s
        """ % (
        str(class_hist.size()),
        str(total_hist.size()),
    )
    assert (
        class_hist > total_hist
    ).sum() == 0, (
        "Invalid. Class histogram must be less than or equal to total histogram"
    )

    num_bins = class_hist.size()[0]
    # Cumsum from highest bucket to lowest
    cum_class_hist = torch.cumsum(torch.flip(class_hist, dims=(0,)), dim=0).double()
    cum_total_hist = torch.cumsum(torch.flip(total_hist, dims=(0,)), dim=0).double()
    class_totals = cum_class_hist[-1, :]

    prec_t = cum_class_hist / cum_total_hist
    recall_t = cum_class_hist / class_totals

    prec = torch.unbind(prec_t, dim=1)
    recall = torch.unbind(recall_t, dim=1)
    assert len(prec) == len(
        recall
    ), "The number of precision curves does not match the number of recall curves"

    final_prec = []
    final_recall = []
    final_ap = []
    for c, prec_curve in enumerate(prec):
        recall_curve = recall[c]
        assert (
            recall_curve.size()[0] == num_bins and prec_curve.size()[0] == num_bins
        ), "Precision and recall curves do not have the correct number of entries"

        # Check if any samples from class were seen
        if class_totals[c] == 0:
            continue

        # Remove duplicate entries
        prev_r = torch.tensor(-1.0).double()
        prev_p = torch.tensor(1.1).double()
        new_recall_curve = torch.tensor([], dtype=torch.double)
        new_prec_curve = torch.tensor([], dtype=torch.double)
        for idx, r in enumerate(recall_curve):
            p = prec_curve[idx]
            # Remove points on PR curve that are invalid
            if r.item() <= 0:
                continue

            # Remove duplicates (due to empty buckets):
            if r.item() == prev_r.item() and p.item() == prev_p.item():
                continue

            # Add points to curve
            new_recall_curve = torch.cat((new_recall_curve, r.unsqueeze(0)), dim=0)
            new_prec_curve = torch.cat((new_prec_curve, p.unsqueeze(0)), dim=0)
            prev_r = r
            prev_p = p

        ap = calc_ap(new_prec_curve, new_recall_curve)
        final_prec.append(new_prec_curve)
        final_recall.append(new_recall_curve)
        final_ap.append(ap)

    return {"prec": final_prec, "recall": final_recall, "ap": final_ap}


def get_checkpoint_dict(task, input_args, deep_copy=False):
    assert isinstance(
        input_args, dict
    ), f"Unexpected input_args of type: {type(input_args)}"
    return {
        "input_args": input_args,
        "classy_state_dict": task.get_classy_state(deep_copy=deep_copy),
    }


def load_checkpoint(
    checkpoint_path: str, device: torch.device = CPU_DEVICE
) -> Optional[Dict]:
    """Loads a checkpoint from the specified checkpoint path.

    Args:
        checkpoint_path: The path to load the checkpoint from. Can be a file or a
            directory. If it is a directory, the checkpoint is loaded from
            :py:data:`CHECKPOINT_FILE` inside the directory.
        device: device to load the checkpoint to

    Returns:
        The checkpoint, if it exists, or None.
    """
    if not checkpoint_path:
        return None

    assert device is not None, "Please specify what device to load checkpoint on"
    assert device.type in ["cpu", "cuda"], f"Unknown device: {device}"
    if device.type == "cuda":
        assert torch.cuda.is_available()

    if not PathManager.exists(checkpoint_path):
        logging.warning(f"Checkpoint path {checkpoint_path} not found")
        return None
    if PathManager.isdir(checkpoint_path):
        checkpoint_path = f"{checkpoint_path.rstrip('/')}/{CHECKPOINT_FILE}"

    if not PathManager.exists(checkpoint_path):
        logging.warning(f"Checkpoint file {checkpoint_path} not found.")
        return None

    logging.info(f"Attempting to load checkpoint from {checkpoint_path}")
    # load model on specified device and not on saved device for model and return
    # the checkpoint
    with PathManager.open(checkpoint_path, "rb") as f:
        checkpoint = torch.load(f, map_location=device)
    logging.info(f"Loaded checkpoint from {checkpoint_path}")
    return checkpoint


def update_classy_model(model, model_state_dict, reset_heads):
    """
    Updates the model with the provided model state dictionary.

    Args:
        model: ClassyVisionModel instance to update
        model_state_dict: State dict, should be the output of a call to
            ClassyVisionModel.get_classy_state().
        reset_heads: if False, uses the heads' state from model_state_dict.
    """
    try:
        if reset_heads:
            current_model_state_dict = model.get_classy_state()
            # replace the checkpointed head states with source head states
            model_state_dict["model"]["heads"] = current_model_state_dict["model"][
                "heads"
            ]
        model.set_classy_state(model_state_dict)
        logging.info("Model state load successful")
        return True
    except Exception:
        logging.exception("Could not load the model state")
    return False


def update_classy_state(task, state_dict):
    """
    Updates the task with the provided task dictionary.

    Args:
        task: ClassyTask instance to update
        state_dict: State dict, should be the output of a call to
            ClassyTask.get_classy_state().
    """
    logging.info("Loading classy state from checkpoint")

    try:
        task.set_classy_state(state_dict)
        logging.info("Checkpoint load successful")
        return True
    except Exception:
        logging.exception("Could not load the checkpoint")

    return False


def save_checkpoint(checkpoint_folder, state, checkpoint_file=CHECKPOINT_FILE):
    """
    Saves a state variable to the specified checkpoint folder. Returns filename
    of checkpoint if successful, and False otherwise.
    """

    # make sure that we have a checkpoint folder:
    if not PathManager.isdir(checkpoint_folder):
        try:
            PathManager.mkdirs(checkpoint_folder)
        except BaseException:
            logging.warning(
                "Could not create folder %s." % checkpoint_folder, exc_info=True
            )
    if not PathManager.isdir(checkpoint_folder):
        return False

    # write checkpoint atomically:
    try:
        full_filename = f"{checkpoint_folder}/{checkpoint_file}"
        with PathManager.open(full_filename, "wb") as f:
            torch.save(state, f)
        return full_filename
    except BaseException:
        logging.warning(
            "Did not write checkpoint to %s." % checkpoint_folder, exc_info=True
        )
        return False


def stepwise_learning_rate(base_lr, epoch, optimizer, epoch_step=30):
    """
    Step-wise reduction of learning rate by a factor of 10 every epoch_step epochs.
    """
    lr = base_lr * (0.1 ** (epoch // epoch_step))
    for param_group in optimizer.param_groups:
        param_group["lr"] = lr


def cosine_learning_rate(base_lr, epoch, optimizer, max_epoch=90):
    """
    Cosine-shaped reduction of learning rate.
    """
    lr = 0.5 * base_lr * (1.0 + math.cos(math.pi * epoch / max_epoch))
    for param_group in optimizer.param_groups:
        param_group["lr"] = lr


def visualize_image(img, mean=None, std=None):
    """
    Make preprocessed image look pretty again to facilitate visualization.
    """

    # assertions:
    assert torch.is_tensor(img) and img.dim() == 3
    for val in [mean, std]:
        if val is not None:
            assert torch.is_tensor(val)
            assert val.dim() == 1 and len(val) == img.size(0)

    # undo mean/std normalization:
    new_img = img.clone()
    if mean is not None and std is not None:
        for c in range(new_img.size(0)):
            new_img[c].mul_(std[c]).add_(mean[c])

    # normalize image to be a byte image:
    if new_img.max() < 1.1:
        new_img.mul_(255.0)
    return new_img.byte()


def set_proxies():
    """Set proxies to allow downloading of external URLs."""
    os.environ["HTTP_PROXY"] = "http://fwdproxy:8080"
    os.environ["HTTPS_PROXY"] = "http://fwdproxy:8080"
    os.environ["http_proxy"] = "fwdproxy:8080"
    os.environ["https_proxy"] = "fwdproxy:8080"


def unset_proxies():
    """Unset proxies to prevent downloading of external URLs."""
    del os.environ["HTTP_PROXY"]
    del os.environ["HTTPS_PROXY"]
    del os.environ["http_proxy"]
    del os.environ["https_proxy"]


def flatten_dict(value_dict, prefix="", sep="_"):
    """
    Flattens nested dict into (key, val) dict. Used for flattening meters
    structure, so that they can be visualized.
    """
    items = []
    for k, v in value_dict.items():
        key = prefix + sep + k if prefix else k
        if isinstance(v, collections.MutableMapping):
            items.extend(flatten_dict(value_dict=v, prefix=key, sep=sep).items())
        else:
            items.append((key, v))
    return dict(items)


def load_json(json_path):
    """
    Loads a json config from a file.
    """
    assert os.path.exists(json_path), "Json file %s not found" % json_path
    json_file = open(json_path)
    json_config = json_file.read()
    json_file.close()
    try:
        config = json.loads(json_config)
    except BaseException as err:
        raise Exception("Failed to validate config with error: %s" % str(err))

    return config


@contextlib.contextmanager
def torch_seed(seed):
    """Context manager which seeds the PyTorch PRNG with the specified seed and
    restores the state afterward. Setting seed to None is equivalent to running
    the code without the context manager."""
    if seed is None:
        yield
        return
    state = torch.get_rng_state()
    torch.manual_seed(seed)
    try:
        yield
    finally:
        torch.set_rng_state(state)


# NOTE [ Python Traceback Reference Cycle Problem ]
#
# When using sys.exc_info(), it is important to **not** store the exc_info[2],
# which is the traceback, because otherwise you will run into the traceback
# reference cycle problem, i.e., the traceback holding reference to the frame,
# and the frame (which holds reference to all the object in its temporary scope)
# holding reference the traceback.
#
# TODO: aadcock: This is a fork of the PyTorch ExceptionWrapper class
# to facilitate the backfill dataloader until we kill it. Once we kill
# datasets/core/backfill_async_dataset_iterator.py we can kill these functions
class KeyErrorMessage(str):
    r"""str subclass that returns itself in repr"""

    def __repr__(self):
        return self


class ExceptionWrapper(object):
    r"""Wraps an exception plus traceback to communicate across threads"""

    def __init__(self, exc_info=None, where="in background"):
        # It is important that we don't store exc_info, see
        # NOTE [ Python Traceback Reference Cycle Problem ]
        if exc_info is None:
            exc_info = sys.exc_info()
        self.exc_type = exc_info[0]
        self.exc_msg = "".join(traceback.format_exception(*exc_info))
        self.where = where

    def reraise(self):
        r"""Reraises the wrapped exception in the current thread"""
        # Format a message such as: "Caught ValueError in DataLoader worker
        # process 2. Original Traceback:", followed by the traceback.
        msg = "Caught {} {}.\nOriginal {}".format(
            self.exc_type.__name__, self.where, self.exc_msg
        )
        if self.exc_type == KeyError:
            # KeyError calls repr() on its argument (usually a dict key). This
            # makes stack traces unreadable. It will not be changed in Python
            # (https://bugs.python.org/issue2651), so we work around it.
            msg = KeyErrorMessage(msg)
        raise self.exc_type(msg)


def convert_to_one_hot(targets, classes):
    """
    This function converts target class indices to one-hot vectors,
    given the number of classes.

    """
    assert (
        torch.max(targets).item() < classes
    ), "Class Index must be less than number of classes"
    one_hot_targets = torch.zeros(
        (targets.shape[0], classes), dtype=torch.long, device=targets.device
    )
    one_hot_targets.scatter_(1, targets.long(), 1)
    return one_hot_targets


def maybe_convert_to_one_hot(target, model_output):
    """
    This function infers whether target is integer or 0/1 encoded
    and converts it to 0/1 encoding if necessary.
    """
    target_shape_list = list(target.size())

    if len(target_shape_list) == 1 or (
        len(target_shape_list) == 2 and target_shape_list[1] == 1
    ):
        target = convert_to_one_hot(target.view(-1, 1), model_output.shape[1])

    assert (target.shape == model_output.shape) and (
        torch.min(target.eq(0) + target.eq(1)) == 1
    ), (
        "Target must be one-hot/multi-label encoded and of the "
        "same shape as model_output."
    )

    return target


def bind_method_to_class(method, cls):
    """
    Binds an already bound method to the provided class.
    """
    return method.__func__.__get__(cls)


def get_model_dummy_input(
    model, input_shape, input_key, batchsize=1, non_blocking=False
):
    if isinstance(input_key, list):
        # av mode, with multiple input keys
        input = {}
        for i, key in enumerate(input_key):
            shape = (batchsize,) + tuple(input_shape[i])
            cur_input = torch.zeros(shape)
            if next(model.parameters()).is_cuda:
                cur_input = cur_input.cuda(non_blocking=non_blocking)
            input[key] = cur_input
    else:
        # add a dimension to represent minibatch axis
        shape = (batchsize,) + tuple(input_shape)
        input = torch.zeros(shape)
        if next(model.parameters()).is_cuda:
            input = input.cuda(non_blocking=non_blocking)
        if input_key:
            input = {input_key: input}
    return input



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from typing import Any, Dict

import torch.optim
from classy_vision.generic.util import is_pos_float

from . import ClassyOptimizer, register_optimizer


@register_optimizer("rmsprop")
class RMSProp(ClassyOptimizer):
    def __init__(
        self,
        lr: float = 0.1,
        momentum: float = 0,
        weight_decay: float = 0,
        alpha: float = 0.99,
        eps: float = 1e-8,
        centered: bool = False,
    ) -> None:
        super().__init__()

        self.parameters.lr = lr
        self.parameters.momentum = momentum
        self.parameters.weight_decay = weight_decay
        self.parameters.alpha = alpha
        self.parameters.eps = eps
        self.parameters.centered = centered

    def init_pytorch_optimizer(self, model, **kwargs):
        super().init_pytorch_optimizer(model, **kwargs)
        self.optimizer = torch.optim.RMSprop(
            self.param_groups_override,
            lr=self.parameters.lr,
            momentum=self.parameters.momentum,
            weight_decay=self.parameters.weight_decay,
            alpha=self.parameters.alpha,
            eps=self.parameters.eps,
            centered=self.parameters.centered,
        )

    @classmethod
    def from_config(cls, config: Dict[str, Any]) -> "RMSProp":
        """Instantiates a RMSProp from a configuration.

        Args:
            config: A configuration for a RMSProp.
                See :func:`__init__` for parameters expected in the config.

        Returns:
            A RMSProp instance.
        """
        # Default params
        config.setdefault("lr", 0.1)
        config.setdefault("momentum", 0.0)
        config.setdefault("weight_decay", 0.0)
        config.setdefault("alpha", 0.99)
        config.setdefault("eps", 1e-8)
        config.setdefault("centered", False)

        for key in ["momentum", "alpha"]:
            assert (
                config[key] >= 0.0 and config[key] < 1.0 and type(config[key]) == float
            ), f"Config must contain a '{key}' in [0, 1) for RMSProp optimizer"
        assert is_pos_float(
            config["eps"]
        ), f"Config must contain a positive 'eps' for RMSProp optimizer"
        assert isinstance(
            config["centered"], bool
        ), "Config must contain a boolean 'centered' param for RMSProp optimizer"

        return cls(
            lr=config["lr"],
            momentum=config["momentum"],
            weight_decay=config["weight_decay"],
            alpha=config["alpha"],
            eps=config["eps"],
            centered=config["centered"],
        )



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import copy
from pathlib import Path

import torch
import torch.nn.modules.loss as torch_losses
from classy_vision.generic.registry_utils import import_all_modules

from .classy_loss import ClassyLoss


FILE_ROOT = Path(__file__).parent


LOSS_REGISTRY = {}
LOSS_CLASS_NAMES = set()


def build_loss(config):
    """Builds a ClassyLoss from a config.

    This assumes a 'name' key in the config which is used to determine what
    model class to instantiate. For instance, a config `{"name": "my_loss",
    "foo": "bar"}` will find a class that was registered as "my_loss"
    (see :func:`register_loss`) and call .from_config on it.

    In addition to losses registered with :func:`register_loss`, we also
    support instantiating losses available in the `torch.nn.modules.loss <https:
    //pytorch.org/docs/stable/nn.html#loss-functions>`_
    module. Any keys in the config will get expanded to parameters of the loss
    constructor. For instance, the following call will instantiate a
    `torch.nn.modules.CrossEntropyLoss <https://pytorch.org/docs/stable/
    nn.html#torch.nn.CrossEntropyLoss>`_:

    .. code-block:: python

     build_loss({"name": "CrossEntropyLoss", "reduction": "sum"})
    """

    assert "name" in config, f"name not provided for loss: {config}"
    name = config["name"]
    if name in LOSS_REGISTRY:
        return LOSS_REGISTRY[name].from_config(config)

    # the name should be available in torch.nn.modules.loss
    assert hasattr(torch_losses, name), (
        f"{name} isn't a registered loss"
        ", nor is it available in torch.nn.modules.loss"
    )
    args = copy.deepcopy(config)
    del args["name"]
    if "weight" in args:
        # if we are passing weights, we need to change the weights from a list
        # to a tensor
        args["weight"] = torch.tensor(args["weight"], dtype=torch.float)
    return getattr(torch_losses, name)(**args)


def register_loss(name):
    """Registers a ClassyLoss subclass.

    This decorator allows Classy Vision to instantiate a subclass of
    ClassyLoss from a configuration file, even if the class itself is not
    part of the Classy Vision framework. To use it, apply this decorator to a
    ClassyLoss subclass, like this:

    .. code-block:: python

     @register_loss("my_loss")
     class MyLoss(ClassyLoss):
          ...

    To instantiate a loss from a configuration file, see
    :func:`build_loss`."""

    def register_loss_cls(cls):
        if name in LOSS_REGISTRY:
            raise ValueError("Cannot register duplicate optimizer ({})".format(name))
        if not issubclass(cls, ClassyLoss):
            raise ValueError(
                "Loss ({}: {}) must extend ClassyLoss".format(name, cls.__name__)
            )
        LOSS_REGISTRY[name] = cls
        LOSS_CLASS_NAMES.add(cls.__name__)
        return cls

    return register_loss_cls


# automatically import any Python files in the losses/ directory
import_all_modules(FILE_ROOT, "classy_vision.losses")


from .barron_loss import BarronLoss  # isort:skip
from .label_smoothing_loss import LabelSmoothingCrossEntropyLoss  # isort:skip
from .multi_output_sum_loss import MultiOutputSumLoss  # isort:skip
from .soft_target_cross_entropy_loss import SoftTargetCrossEntropyLoss  # isort:skip
from .sum_arbitrary_loss import SumArbitraryLoss  # isort:skip


__all__ = [
    "BarronLoss",
    "ClassyLoss",
    "LabelSmoothingCrossEntropyLoss",
    "MultiOutputSumLoss",
    "SoftTargetCrossEntropyLoss",
    "SumArbitraryLoss",
    "build_loss",
    "register_loss",
]



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import logging
import time
from typing import Any, Dict, List, Optional

from classy_vision import tasks
from classy_vision.generic.distributed_util import is_master
from classy_vision.hooks.classy_hook import ClassyHook


try:
    from tensorboardX import SummaryWriter  # noqa F401

    tbx_available = True
except ImportError:
    tbx_available = False


log = logging.getLogger()


class TensorboardPlotHook(ClassyHook):
    """
    Hook for writing the losses, learning rates and meters to `tensorboard <https
    ://www.tensorflow.org/tensorboard`>_.

    Global steps are counted in terms of the number of samples processed.
    """

    on_start = ClassyHook._noop
    on_forward = ClassyHook._noop
    on_loss_and_meter = ClassyHook._noop
    on_end = ClassyHook._noop

    def __init__(self, tb_writer) -> None:
        """The constructor method of TensorboardPlotHook.

        Args:
            tb_writer: `Tensorboard SummaryWriter <https://tensorboardx.
            readthedocs.io/en/latest/tensorboard.html#tensorboardX.
            SummaryWriter>`_ instance
        """
        super().__init__()
        if not tbx_available:
            raise RuntimeError(
                "tensorboardX not installed, cannot use TensorboardPlotHook"
            )

        self.tb_writer = tb_writer
        self.learning_rates: Optional[List[float]] = None
        self.wall_times: Optional[List[float]] = None
        self.num_steps_global: Optional[List[int]] = None

    def on_phase_start(
        self, task: "tasks.ClassyTask", local_variables: Dict[str, Any]
    ) -> None:
        """Initialize losses and learning_rates."""
        self.learning_rates = []
        self.wall_times = []
        self.num_steps_global = []

    def on_step(
        self, task: "tasks.ClassyTask", local_variables: Dict[str, Any]
    ) -> None:
        """Store the observed learning rates."""
        if self.learning_rates is None:
            logging.warning("learning_rates is not initialized")
            return

        if not task.train:
            # Only need to log the average loss during the test phase
            return

        learning_rate_val = task.optimizer.parameters.lr

        self.learning_rates.append(learning_rate_val)
        self.wall_times.append(time.time())
        self.num_steps_global.append(task.num_updates)

    def on_phase_end(
        self, task: "tasks.ClassyTask", local_variables: Dict[str, Any]
    ) -> None:
        """Add the losses and learning rates to tensorboard."""
        if self.learning_rates is None:
            logging.warning("learning_rates is not initialized")
            return

        batches = len(task.losses)
        if batches == 0 or not is_master():
            return

        phase_type = task.phase_type
        phase_type_idx = task.train_phase_idx if task.train else task.eval_phase_idx

        logging.info(f"Plotting to Tensorboard for {phase_type} phase {phase_type_idx}")

        phase_type = task.phase_type
        loss_key = f"{phase_type}_loss"
        learning_rate_key = f"{phase_type}_learning_rate_updates"

        if task.train:
            for loss, learning_rate, global_step, wall_time in zip(
                task.losses, self.learning_rates, self.num_steps_global, self.wall_times
            ):
                loss /= task.get_batchsize_per_replica()
                self.tb_writer.add_scalar(
                    loss_key, loss, global_step=global_step, walltime=wall_time
                )
                self.tb_writer.add_scalar(
                    learning_rate_key,
                    learning_rate,
                    global_step=global_step,
                    walltime=wall_time,
                )

        loss_avg = sum(task.losses) / (batches * task.get_batchsize_per_replica())

        loss_key = "avg_{phase_type}_loss".format(phase_type=task.phase_type)
        self.tb_writer.add_scalar(loss_key, loss_avg, global_step=phase_type_idx)

        # plot meters which return a dict
        for meter in task.meters:
            if not isinstance(meter.value, dict):
                log.warn(f"Skipping meter {meter.name} with value: {meter.value}")
                continue
            for name, value in meter.value.items():
                if isinstance(value, float):
                    meter_key = f"{phase_type}_{meter.name}_{name}"
                    self.tb_writer.add_scalar(
                        meter_key, value, global_step=phase_type_idx
                    )
                else:
                    log.warn(
                        f"Skipping meter name {meter.name}_{name} with value: {value}"
                    )
                    continue

        if hasattr(task, "perf_log"):
            for perf in task.perf_log:
                phase_idx = perf["phase_idx"]
                tag = perf["tag"]
                for metric_name, metric_value in perf.items():
                    if metric_name in ["phase_idx", "tag"]:
                        continue

                    self.tb_writer.add_scalar(
                        f"Speed/{tag}/{metric_name}",
                        metric_value,
                        global_step=phase_idx,
                    )

        # flush so that the plots aren't lost if training crashes soon after
        self.tb_writer.flush()
        logging.info(f"Done plotting to Tensorboard")



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import copy
import random
import unittest

import numpy
import torch
import torchvision.transforms as transforms
from classy_vision.dataset.core.random_image_datasets import (
    RandomImageBinaryClassDataset,
    SampleType,
)
from classy_vision.dataset.transforms import build_transforms
from classy_vision.dataset.transforms.util import (
    GenericImageTransform,
    ImagenetAugmentTransform,
    ImagenetNoAugmentTransform,
    build_field_transform_default_imagenet,
)


def _apply_transform_to_key_and_copy(sample, transform, key, seed=0):
    """
    This helper function takes a sample, makes a copy, applies the
    provided transform to the appropriate key in the copied sample and
    returns the copy. It's solely to help make sure the copying /
    random seed happens correctly throughout the file.

    It is useful for constructing the expected sample field in the
    transform checks.
    """
    expected_sample = copy.deepcopy(sample)
    torch.manual_seed(seed)
    numpy.random.seed(seed)
    random.seed(seed)
    is_tuple = False
    if isinstance(expected_sample, tuple):
        expected_sample = list(expected_sample)
        is_tuple = True
    expected_sample[key] = transform(expected_sample[key])
    return tuple(expected_sample) if is_tuple else expected_sample


class DatasetTransformsUtilTest(unittest.TestCase):
    def get_test_image_dataset(self, sample_type):
        return RandomImageBinaryClassDataset(
            crop_size=224,
            class_ratio=0.5,
            num_samples=100,
            seed=0,
            sample_type=sample_type,
        )

    def transform_checks(self, sample, transform, expected_sample, seed=0):
        """
        This helper function applies the transform to the sample
        and verifies that the output is the expected_sample. The
        sole purpose is to make sure copying / random seed / checking
        all of the fields in the sample happens correctly.
        """
        transformed_sample = copy.deepcopy(sample)
        torch.manual_seed(seed)
        numpy.random.seed(seed)
        random.seed(seed)
        transformed_sample = transform(transformed_sample)

        if isinstance(expected_sample, (tuple, list)):
            for transformed, exp in zip(transformed_sample, expected_sample):
                if torch.is_tensor(exp):
                    self.assertTrue(torch.allclose(transformed, exp))

        if isinstance(expected_sample, dict):
            for key, exp_val in expected_sample.items():
                self.assertTrue(key in transformed_sample)
                if torch.is_tensor(exp_val):
                    self.assertTrue(torch.allclose(transformed_sample[key], exp_val))
                elif isinstance(exp_val, float):
                    self.assertAlmostEqual(transformed_sample[key], exp_val)
                else:
                    self.assertEqual(transformed_sample[key], exp_val)

    def test_build_dict_field_transform_default_imagenet(self):
        dataset = self.get_test_image_dataset(SampleType.DICT)

        # should apply the transform in the config
        config = [{"name": "ToTensor"}]
        default_transform = transforms.Compose(
            [transforms.CenterCrop(100), transforms.ToTensor()]
        )
        transform = build_field_transform_default_imagenet(
            config, default_transform=default_transform
        )
        sample = dataset[0]
        expected_sample = _apply_transform_to_key_and_copy(
            sample, transforms.ToTensor(), "input"
        )
        self.transform_checks(sample, transform, expected_sample)

        # should apply default_transform
        config = None
        transform = build_field_transform_default_imagenet(
            config, default_transform=default_transform
        )
        expected_sample = _apply_transform_to_key_and_copy(
            sample, default_transform, "input"
        )
        self.transform_checks(sample, transform, expected_sample)

        # should apply the transform for a test split
        transform = build_field_transform_default_imagenet(config, split="test")
        expected_sample = _apply_transform_to_key_and_copy(
            sample, ImagenetNoAugmentTransform(), "input"
        )
        self.transform_checks(sample, transform, expected_sample)

    def test_build_tuple_field_transform_default_imagenet(self):
        dataset = self.get_test_image_dataset(SampleType.TUPLE)

        # should apply the transform in the config
        config = [{"name": "ToTensor"}]
        default_transform = transforms.Compose(
            [transforms.CenterCrop(100), transforms.ToTensor()]
        )
        transform = build_field_transform_default_imagenet(
            config, default_transform=default_transform, key=0, key_map_transform=None
        )
        sample = dataset[0]
        expected_sample = _apply_transform_to_key_and_copy(
            sample, transforms.ToTensor(), 0
        )
        self.transform_checks(sample, transform, expected_sample)

        # should apply default_transform
        config = None
        transform = build_field_transform_default_imagenet(
            config, default_transform=default_transform, key=0, key_map_transform=None
        )
        sample = dataset[0]
        expected_sample = _apply_transform_to_key_and_copy(sample, default_transform, 0)
        self.transform_checks(sample, transform, expected_sample)

        # should apply the transform for a test split
        transform = build_field_transform_default_imagenet(
            config, split="test", key=0, key_map_transform=None
        )
        sample = dataset[0]
        expected_sample = _apply_transform_to_key_and_copy(
            sample, ImagenetNoAugmentTransform(), 0
        )
        self.transform_checks(sample, transform, expected_sample)

    def test_apply_transform_to_key_from_config(self):
        dataset = self.get_test_image_dataset(SampleType.DICT)

        config = [
            {
                "name": "apply_transform_to_key",
                "transforms": [{"name": "ToTensor"}],
                "key": "input",
            }
        ]
        transform = build_transforms(config)
        sample = dataset[0]
        expected_sample = _apply_transform_to_key_and_copy(
            sample, transforms.ToTensor(), "input"
        )
        self.transform_checks(sample, transform, expected_sample)

    def test_generic_image_transform(self):
        dataset = self.get_test_image_dataset(SampleType.TUPLE)

        # Check constructor asserts
        with self.assertRaises(AssertionError):
            transform = GenericImageTransform(
                split="train", transform=transforms.ToTensor()
            )
            transform = GenericImageTransform(split="valid", transform=None)

        # Check class constructor
        transform = GenericImageTransform(transform=None)
        PIL_sample = dataset[0]
        tensor_sample = (transforms.ToTensor()(PIL_sample[0]), PIL_sample[1])
        expected_sample = {
            "input": copy.deepcopy(tensor_sample[0]),
            "target": copy.deepcopy(tensor_sample[1]),
        }
        self.transform_checks(tensor_sample, transform, expected_sample)

        transform = GenericImageTransform(transform=transforms.ToTensor())
        sample = dataset[0]
        expected_sample = _apply_transform_to_key_and_copy(
            {"input": sample[0], "target": sample[1]}, transforms.ToTensor(), "input"
        )
        self.transform_checks(sample, transform, expected_sample)

        transform = GenericImageTransform(split="train")
        sample = dataset[0]
        expected_sample = _apply_transform_to_key_and_copy(
            {"input": sample[0], "target": sample[1]},
            ImagenetAugmentTransform(),
            "input",
        )
        self.transform_checks(sample, transform, expected_sample)

        transform = GenericImageTransform(split="test")
        sample = dataset[0]
        expected_sample = _apply_transform_to_key_and_copy(
            {"input": sample[0], "target": sample[1]},
            ImagenetNoAugmentTransform(),
            "input",
        )
        self.transform_checks(sample, transform, expected_sample)

        # Check from_config constructor / registry
        config = [
            {"name": "generic_image_transform", "transforms": [{"name": "ToTensor"}]}
        ]
        transform = build_transforms(config)
        sample = dataset[0]
        expected_sample = _apply_transform_to_key_and_copy(
            {"input": sample[0], "target": sample[1]}, transforms.ToTensor(), "input"
        )
        self.transform_checks(sample, transform, expected_sample)

        # Check with Imagenet defaults
        config = [{"name": "generic_image_transform", "split": "train"}]
        transform = build_transforms(config)
        sample = dataset[0]
        expected_sample = _apply_transform_to_key_and_copy(
            {"input": sample[0], "target": sample[1]},
            ImagenetAugmentTransform(),
            "input",
        )
        self.transform_checks(sample, transform, expected_sample)

        config = [{"name": "generic_image_transform", "split": "test"}]
        transform = build_transforms(config)
        sample = dataset[0]
        expected_sample = _apply_transform_to_key_and_copy(
            {"input": sample[0], "target": sample[1]},
            ImagenetNoAugmentTransform(),
            "input",
        )
        self.transform_checks(sample, transform, expected_sample)



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import logging
import os
from typing import Any, Callable, Dict, Optional

import torch
from classy_vision.generic.distributed_util import get_rank, get_world_size
from torch.utils.data import Sampler
from torchvision import get_video_backend
from torchvision.datasets.samplers.clip_sampler import (
    DistributedSampler,
    RandomClipSampler,
    UniformClipSampler,
)

from .classy_dataset import ClassyDataset


class MaxLengthClipSampler(Sampler):
    """MaxLengthClipSampler is a thin wrapper on top of clip samplers in TorchVision.

    It takes as input a TorchVision clip sampler, and an optional argument
    `num_samples` to limit the number of samples.
    """

    def __init__(self, clip_sampler, num_samples=None):
        """The constructor method of MaxLengthClipSampler.

        Args:
            clip_sampler: clip sampler without a limit on the total number of clips
                it can sample, such as RandomClipSampler and UniformClipSampler.
            num_samples: if provided, it denotes the maximal number of clips the sampler
                will return

        """
        self.clip_sampler = clip_sampler
        self.num_samples = num_samples

    def __iter__(self):
        num_samples = len(self)
        n = 0
        for clip in self.clip_sampler:
            if n < num_samples:
                yield clip
                n += 1
            else:
                break

    def __len__(self):
        full_size = len(self.clip_sampler)
        if self.num_samples is None:
            return full_size

        return min(full_size, self.num_samples)


class ClassyVideoDataset(ClassyDataset):
    """Interface specifying what a ClassyVision video dataset is expected to provide.

    This dataset considers every video as a collection of video clips of fixed size,
    specified by ``frames_per_clip``, where the step in frames between each clip
    is given by ``step_between_clips``. It uses a clip sampler to sample
    a specified number of clips (``clips_per_video``) from each video.
    For training set, a random clip sampler is used to
    sample a small number of clips (e.g. 1) from each video
    For testing set, a uniform clip sampler is used to evenly sample a large
    number of clips (e.g. 10) from the video.

    To give an example, for 2 videos with 10 and 15 frames respectively,
    if ``frames_per_clip=5`` and ``step_between_clips=5``, the dataset size
    will be (2 + 3) = 5, where the first two elements will come from video 1,
    and the next three elements from video 2. Note that we drop clips which do
    not have exactly ``frames_per_clip`` elements, so not all frames in a video
    may be present.
    """

    def __init__(
        self,
        dataset: Any,
        split: str,
        batchsize_per_replica: int,
        shuffle: bool,
        transform: Callable,
        num_samples: Optional[int],
        clips_per_video: int,
    ):
        """The constructor method of ClassyVideoDataset.

        Args:
            dataset: the underlying video dataset from either TorchVision or other
                source. It should have an attribute *video_clips* of type
                `torchvision.datasets.video_utils.VideoClips <https://github.com/
                pytorch/vision/blob/master/torchvision/datasets/
                video_utils.py#L46/>`_
            split: dataset split. Must be either "train" or "test"
            batchsize_per_replica: batch size per model replica
            shuffle: If true, shuffle video clips.
            transform: callable function to transform video clip sample from
                ClassyVideoDataset
            num_samples: If provided, return at most `num_samples` video clips
            clips_per_video: The number of clips sampled from each video

        """
        super(ClassyVideoDataset, self).__init__(
            dataset, batchsize_per_replica, shuffle, transform, num_samples
        )
        # Assignments:
        self.clips_per_video = clips_per_video
        self.split = split
        self.video_backend = get_video_backend()

    @classmethod
    def parse_config(cls, config: Dict[str, Any]):
        """Parse config to prepare arguments needed by the class constructor."""
        assert "frames_per_clip" in config, "frames_per_clip must be set"

        video_width = config.get("video_width", 0)
        video_height = config.get("video_height", 0)
        video_min_dimension = config.get("video_min_dimension", 0)
        audio_samples = config.get("audio_samples", 0)
        step_between_clips = config.get("step_between_clips", 1)
        frame_rate = config.get("frame_rate", None)
        clips_per_video = config.get("clips_per_video", 1)

        (
            transform_config,
            batchsize_per_replica,
            shuffle,
            num_samples,
        ) = super().parse_config(config)

        if not config["split"] == "train":
            # At testing time, we do not crop frames but conduct a FCN-style evaluation.
            # Video spatial resolution can vary from video to video. So we test one
            # video at a time, and NO. of clips in a minibatch should be equal to
            # No. of clips sampled from a video
            if not batchsize_per_replica == clips_per_video:
                logging.warning(
                    f"For testing, batchsize per replica ({batchsize_per_replica})"
                    + f"should be equal to clips_per_video ({clips_per_video})"
                )

        return (
            transform_config,
            batchsize_per_replica,
            shuffle,
            num_samples,
            config["frames_per_clip"],
            video_width,
            video_height,
            video_min_dimension,
            audio_samples,
            step_between_clips,
            frame_rate,
            clips_per_video,
        )

    @classmethod
    def load_metadata(
        cls,
        filepath: str,
        video_dir: Optional[str] = None,
        update_file_path: bool = False,
    ) -> Dict[str, Any]:
        """Load pre-computed video dataset meta data.

        Video dataset meta data computation takes minutes on small dataset and hours
        on large dataset, and thus is time-consuming. However, it only needs to be
        computed once, and can be saved into a file via :func:`save_metadata`.

        The format of meta data is defined in `TorchVision <https://github.com/
        pytorch/vision/blob/master/torchvision/datasets/video_utils.py#L131/>`_.

        For each video, meta data contains the video file path, presentation
        timestamps of all video frames, and video fps.

        Args:
            filepath: file path of pre-computed meta data
            video_dir: If provided, the folder where video files are stored.
            update_file_path: If true, replace the directory part of video file path
                in meta data with the actual video directory provided in `video_dir`.
                This is necessary for successsfully reusing pre-computed meta data
                when video directory has been moved and is no longer consitent
                with the full video file path saved in the meta data.
        """
        metadata = torch.load(filepath)
        if video_dir is not None and update_file_path:
            # video path in meta data can be computed in a different root video folder
            # If we use a different root video folder, we need to update the video paths
            assert os.path.exists(video_dir), "folder does not exist: %s" % video_dir
            for idx, video_path in enumerate(metadata["video_paths"]):
                # video path template is $VIDEO_DIR/$CLASS_NAME/$VIDEO_FILE
                dirname, filename = os.path.split(video_path)
                _, class_name = os.path.split(dirname)
                metadata["video_paths"][idx] = os.path.join(
                    video_dir, class_name, filename
                )
        return metadata

    @classmethod
    def save_metadata(cls, metadata: Dict[str, Any], filepath: str):
        """Save dataset meta data into a file.

        Args:
            metadata: dataset meta data, which contains video meta infomration, such
                as video file path, video fps, video frame timestamp in each video.
                For the format of dataset meta data, check the `TorchVision
                documentation <https://github.com/pytorch/vision/blob/master/
                torchvision/datasets/video_utils.py#L132-L137/>`_.
            filepath: file path where the meta data will be saved

        """
        filedir = os.path.dirname(filepath)
        if not os.path.exists(filedir):
            try:
                os.mkdirs(filedir)
            except Exception as err:
                logging.warn(f"Fail to create folder: {filedir}")
                raise err
        logging.info(f"Save metadata to file: {filedir}")
        try:
            torch.save(metadata, filepath)
        except ValueError:
            logging.warn(f"Fail to save metadata to file: {filepath}")

    @property
    def video_clips(self):
        """Attribute video_clips.

        It is used in ``_get_sampler`` method. Its data type should be
            `torchvision.datasets.video_utils.VideoClips <https://github.com/
            pytorch/vision/blob/master/torchvision/datasets/video_utils.py#L46/>`_.
        """
        return self.dataset.video_clips

    def _get_sampler(self, epoch) -> "DistributedSampler":
        if self.split == "train":
            # For video model training, we don't necessarily want to use all possible
            # clips in the video in one training epoch. More often, we randomly
            # sample at most N clips per training video. In practice, N is often 1
            clip_sampler = RandomClipSampler(self.video_clips, self.clips_per_video)
        else:
            # For video model testing, we sample N evenly spaced clips per test
            # video. We will simply average predictions over them
            clip_sampler = UniformClipSampler(self.video_clips, self.clips_per_video)
        clip_sampler = MaxLengthClipSampler(clip_sampler, num_samples=self.num_samples)
        world_size = get_world_size()
        rank = get_rank()
        sampler = DistributedSampler(
            clip_sampler,
            num_replicas=world_size,
            rank=rank,
            shuffle=self.shuffle,
            group_size=self.clips_per_video,
        )
        sampler.set_epoch(epoch)
        return sampler

    def _worker_init_fn(self, worker_id):
        # we need to set video backend in the worker process explicitly
        # because the global variable `_video_backend` in TorchVision will
        # always start with the default value `pyav` when multiprocessing
        # context other than `fork` is used, and it won't inherit the value of
        # `_video_backend` in the main process

        from torchvision import set_video_backend

        set_video_backend(self.video_backend)

    def iterator(self, *args, **kwargs):
        kwargs["worker_init_fn"] = self._worker_init_fn
        return super(ClassyVideoDataset, self).iterator(*args, **kwargs)



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import copy
from collections import defaultdict
from pathlib import Path

from classy_vision.generic.registry_utils import import_all_modules
from classy_vision.heads import build_head

from .classy_model import ClassyModel


FILE_ROOT = Path(__file__).parent


MODEL_REGISTRY = {}
MODEL_CLASS_NAMES = set()


def register_model(name):
    """Registers a :class:`ClassyModel` subclass.

    This decorator allows Classy Vision to instantiate a subclass of
    :class:`ClassyModel` from a configuration file, even if the class itself is
    not part of the Classy Vision framework. To use it, apply this decorator to
    a ClassyModel subclass, like this:

    .. code-block:: python

      @register_model('resnet')
      class ResidualNet(ClassyModel):
         ...

    To instantiate a model from a configuration file, see
    :func:`build_model`."""

    def register_model_cls(cls):
        if name in MODEL_REGISTRY:
            raise ValueError("Cannot register duplicate model ({})".format(name))
        if not issubclass(cls, ClassyModel):
            raise ValueError(
                "Model ({}: {}) must extend ClassyModel".format(name, cls.__name__)
            )
        if cls.__name__ in MODEL_CLASS_NAMES:
            raise ValueError(
                "Cannot register model with duplicate class name ({})".format(
                    cls.__name__
                )
            )
        MODEL_REGISTRY[name] = cls
        MODEL_CLASS_NAMES.add(cls.__name__)
        return cls

    return register_model_cls


def build_model(config):
    """Builds a ClassyModel from a config.

    This assumes a 'name' key in the config which is used to determine what
    model class to instantiate. For instance, a config `{"name": "my_model",
    "foo": "bar"}` will find a class that was registered as "my_model"
    (see :func:`register_model`) and call .from_config on it."""

    assert config["name"] in MODEL_REGISTRY, "unknown model"
    model = MODEL_REGISTRY[config["name"]].from_config(config)
    if "heads" in config:
        heads = defaultdict(dict)
        for head_config in config["heads"]:
            assert "fork_block" in head_config, "Expect fork_block in config"
            fork_block = head_config["fork_block"]
            updated_config = copy.deepcopy(head_config)
            del updated_config["fork_block"]

            head = build_head(updated_config)
            heads[fork_block][head.unique_id] = head
        model.set_heads(heads)
    return model


# automatically import any Python files in the models/ directory
import_all_modules(FILE_ROOT, "classy_vision.models")

from .classy_block import ClassyBlock  # isort:skip
from .classy_model import ClassyModelEvaluationMode  # isort:skip
from .classy_model_wrapper import ClassyModelWrapper  # isort:skip
from .densenet import DenseNet  # isort:skip
from .mlp import MLP  # isort:skip
from .resnet import ResNet  # isort:skip
from .resnext import ResNeXt  # isort:skip
from .resnext3d import ResNeXt3D  # isort:skip


__all__ = [
    "build_model",
    "register_model",
    "ClassyBlock",
    "ClassyModel",
    "ClassyModelEvaluationMode",
    "ClassyModelWrapper",
    "DenseNet",
    "MLP",
    "ResNet",
    "ResNeXt",
    "ResNeXt3D",
]



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.


from abc import ABC, abstractmethod
from typing import Any, Dict, Optional


class ClassyTask(ABC):
    """
    An abstract base class for a training task.

    A ClassyTask encapsulates all the components and steps needed
    to train using a :class:`classy_vision.trainer.ClassyTrainer`.
    """

    def __init__(self) -> "ClassyTask":
        """
        Constructs a ClassyTask.
        """
        self.hooks = []

    @classmethod
    @abstractmethod
    def from_config(cls, config: Dict[str, Any]) -> "ClassyTask":
        """Instantiates a ClassyTask from a configuration.

        Args:
            config: A configuration for a ClassyTask.

        Returns:
            A ClassyTask instance.
        """
        raise NotImplementedError()

    @property
    @abstractmethod
    def where(self) -> float:
        """
        Tells how far along (where) we are during training.

        Returns:
            A float in [0, 1) which tells the training progress.
        """
        pass

    @abstractmethod
    def done_training(self) -> bool:
        """
        Tells if we are done training.

        Returns:
            A boolean telling if training is over.
        """
        pass

    @abstractmethod
    def get_classy_state(self, deep_copy: bool = False) -> Dict[str, Any]:
        """Get the state of the ClassyTask.

        The returned state is used for checkpointing.

        Args:
            deep_copy: If True, creates a deep copy of the state dict. Otherwise, the
                returned dict's state will be tied to the object's.

        Returns:
            A state dictionary containing the state of the task.
        """
        pass

    @abstractmethod
    def set_classy_state(self, state):
        """Set the state of the ClassyTask.

        Args:
            state_dict: The state dictionary. Must be the output of a call to
                :func:`get_classy_state`.

        This is used to load the state of the task from a checkpoint.
        """
        pass

    @abstractmethod
    def prepare(
        self,
        num_dataloader_workers=0,
        pin_memory=False,
        use_gpu=False,
        dataloader_mp_context=None,
    ) -> None:
        """
        Prepares the task for training.

        Will be called by the :class:`classy_vision.trainer.ClassyTrainer` to
        prepare the task, before on_start is called.

        Args:
            num_dataloader_workers: Number of workers to create for the dataloaders
            pin_memory: Whether the dataloaders should copy the Tensors into CUDA
                pinned memory (default False)
            use_gpu: True if training on GPUs, False otherwise
        """
        pass

    @abstractmethod
    def train_step(self, use_gpu, local_variables: Optional[Dict] = None) -> None:
        """
        Run a train step.

        This corresponds to training over one batch of data from the dataloaders.

        Args:
            use_gpu: True if training on GPUs, False otherwise
            local_variables: Local variables created in the function. Can be passed to
                custom :class:`classy_vision.hooks.ClassyHook`.
        """
        pass

    @abstractmethod
    def on_start(self, local_variables):
        """
        Start training.

        Called by :class:`classy_vision.trainer.ClassyTrainer` before training starts.
        """
        pass

    @abstractmethod
    def on_phase_start(self, local_variables):
        """
        Epoch start.

        Called by :class:`classy_vision.trainer.ClassyTrainer` before each epoch starts.
        """
        pass

    @abstractmethod
    def on_phase_end(self, local_variables):
        """
        Epoch end.

        Called by :class:`classy_vision.trainer.ClassyTrainer` after each epoch ends.
        """
        pass

    @abstractmethod
    def on_end(self, local_variables):
        """
        Training end.

        Called by :class:`classy_vision.trainer.ClassyTrainer` after training ends.
        """
        pass

    @abstractmethod
    def eval_step(self, use_gpu, local_variables: Optional[Dict] = None) -> None:
        """
        Run an evaluation step.

        This corresponds to evaluating the model over one batch of data.

        Args:
            use_gpu: True if training on GPUs, False otherwise
            local_variables: Local variables created in the function. Can be passed to
                custom :class:`classy_vision.hooks.ClassyHook`.
        """
        pass

    def step(self, use_gpu, local_variables: Optional[Dict] = None) -> None:
        from classy_vision.hooks import ClassyHookFunctions

        if self.train:
            self.train_step(use_gpu, local_variables)
        else:
            self.eval_step(use_gpu, local_variables)

        self.run_hooks(local_variables, ClassyHookFunctions.on_step.name)

    def run_hooks(self, local_variables: Dict[str, Any], hook_function: str) -> None:
        """
        Helper function that runs a hook function for all the
        :class:`classy_vision.hooks.ClassyHook`.

        Args:
            local_variables: Local variables created in :func:`train_step`
            hook_function: One of the hook functions in the
            :class:`classy_vision.hooks.ClassyHookFunctions`
                enum.
        """
        for hook in self.hooks:
            getattr(hook, hook_function)(self, local_variables)



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import unittest

import torch
from classy_vision.heads import ClassyHead, build_head, register_head


class TestClassyHead(unittest.TestCase):
    @register_head("dummy_head")
    class DummyHead(ClassyHead):
        def __init__(self, unique_id, num_classes, in_plane):
            super().__init__(unique_id, num_classes)
            self.fc = torch.nn.Linear(in_plane, num_classes)

        def forward(self, x):
            return self.fc(x)

        @classmethod
        def from_config(cls, config):
            return cls(config["unique_id"], config["num_classes"], config["in_plane"])

    def _get_config(self):
        return {
            "name": "dummy_head",
            "num_classes": 3,
            "unique_id": "cortex_dummy_head",
            "fork_block": "block3",
            "in_plane": 2048,
        }

    def test_build_head(self):
        config = self._get_config()
        head = build_head(config)
        self.assertEqual(head.unique_id, config["unique_id"])

        del config["unique_id"]
        with self.assertRaises(AssertionError):
            head = build_head(config)

    def test_forward(self):
        config = self._get_config()
        head = build_head(config)
        input = torch.randn(1, config["in_plane"])
        output = head(input)
        self.assertEqual(output.size(), torch.Size([1, 3]))

    def _get_pass_through_config(self):
        return {
            "name": "identity",
            "num_classes": 3,
            "unique_id": "cortex_pass_through_head",
            "fork_block": "block3",
            "in_plane": 4,
        }

    def test_identity_forward(self):
        config = self._get_pass_through_config()
        head = build_head(config)
        input = torch.randn(1, config["in_plane"])
        output = head(input)
        self.assertEqual(input.size(), output.size())
        self.assert_(torch.all(torch.eq(input, output)))



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import os.path
from typing import Any, Callable, Dict, List, Optional, Union

import torch
import torchvision.datasets as datasets
import torchvision.transforms as transforms

from .classy_dataset import ClassyDataset
from .core import ListDataset
from .transforms import build_transforms
from .transforms.classy_transform import ClassyTransform
from .transforms.util import TupleToMapTransform


def _load_dataset(image_paths, targets):
    if targets is None:
        targets = [torch.tensor([]) for _ in image_paths]
    if isinstance(image_paths, str):
        assert os.path.isdir(
            image_paths
        ), "Expect image_paths to be a dir when it is a string"
        dataset = datasets.ImageFolder(image_paths)
        preproc_transform = TupleToMapTransform(list_of_map_keys=["input", "target"])
    else:
        dataset = ListDataset(image_paths, targets)
        preproc_transform = None
    return dataset, preproc_transform


class ImagePathDataset(ClassyDataset):
    """Dataset which reads images from a local filesystem. Implements ClassyDataset.

    The image paths provided can be:
        - A single directory location, in which case the data is expected to be
            arranged in a format similar to :class:`torchvision.datasets.ImageFolder`.
            The targets will be inferred from the directory structure.
        - A list of paths, in which case the list will contain the paths to all the
            images. In this situation, the targets can be specified by the targets
            argument.
    """

    def __init__(
        self,
        batchsize_per_replica: int,
        shuffle: bool,
        transform: Optional[Union[ClassyTransform, Callable]],
        num_samples: Optional[int],
        image_paths: Union[str, List[str]],
        targets: Optional[List[Any]] = None,
    ):
        """Constructor for ImagePathDataset.

        Args:
            batchsize_per_replica: Positive integer indicating batch size for each
                replica
            shuffle: Whether we should shuffle between epochs
            transform: Transform to be applied to each sample
            num_samples: When set, this restricts the number of samples provided by
                the dataset
            image_paths: A directory or a list of file paths where images can be found.
            targets: If a list of file paths is specified, this argument can
                be used to specify a target for each path (must be same length
                as list of file paths). If no targets are needed or image_paths is
                a directory, then targets should be None.
        """
        # TODO(@mannatsingh): we should be able to call build_dataset() to create
        # datasets from this class.
        assert image_paths is not None, "image_paths needs to be provided"
        assert targets is None or isinstance(image_paths, list), (
            "targets cannot be specified when image_paths is a directory containing "
            "the targets in the directory structure"
        )
        dataset, preproc_transform = _load_dataset(image_paths, targets)
        super().__init__(
            dataset, batchsize_per_replica, shuffle, transform, num_samples
        )
        # Some of the base datasets from _load_dataset have different
        # sample formats, the preproc_transform should map them all to
        # the dict {"input": img, "target": label} format
        if preproc_transform is not None:
            self.transform = transforms.Compose([preproc_transform, self.transform])

    @classmethod
    def from_config(
        cls,
        config: Dict[str, Any],
        image_paths: Union[str, List[str]],
        targets: Optional[List[Any]] = None,
    ):
        """Instantiates ImagePathDataset from a config.

        Because image_paths / targets can be arbitrarily long, we
        allow passing in the image paths and targets from python in
        addition to the configuration parameter.

        Args:
            config: A configuration for ImagePathDataset.
                See :func:`__init__` for parameters expected in the config.
            image_paths: Directory or list of image paths.
                See :func:`__init__` for more details
            targets: Optional list of targets for dataset.
                See :func:`__init__` for more details
        """
        (
            transform_config,
            batchsize_per_replica,
            shuffle,
            num_samples,
        ) = cls.parse_config(config)

        transform = build_transforms(transform_config)
        return cls(
            batchsize_per_replica,
            shuffle,
            transform,
            num_samples,
            image_paths,
            targets=targets,
        )



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from abc import ABC, abstractmethod
from typing import Any, Dict


class ClassyTransform(ABC):
    """
    Class representing a data transform abstraction.

    Data transform is most often needed to pre-process input data (e.g. image, video)
    before sending it to a model. But it can also be used for other purposes.
    """

    @abstractmethod
    def __call__(self, image):
        """
        The interface `__call__` is used to transform the input data. It should contain
        the actual implementation of data transform.

        Args:
            image: input image data
        """
        pass

    @classmethod
    def from_config(cls, config: Dict[str, Any]):
        return cls(**config)



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from typing import Any, Dict

from classy_vision.generic.util import load_checkpoint, update_classy_model
from classy_vision.tasks import ClassificationTask, register_task


@register_task("fine_tuning")
class FineTuningTask(ClassificationTask):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.pretrained_checkpoint = None
        self.reset_heads = False
        self.freeze_trunk = False

    @classmethod
    def from_config(cls, config: Dict[str, Any]) -> "FineTuningTask":
        """Instantiates a FineTuningTask from a configuration.

        Args:
            config: A configuration for a FineTuningTask.
                See :func:`__init__` for parameters expected in the config.

        Returns:
            A FineTuningTask instance.
        """
        task = super().from_config(config)

        pretrained_checkpoint = load_checkpoint(config.get("pretrained_checkpoint"))

        if pretrained_checkpoint is not None:
            task.set_pretrained_checkpoint(pretrained_checkpoint)

        task.set_reset_heads(config.get("reset_heads", False))
        task.set_freeze_trunk(config.get("freeze_trunk", False))
        return task

    def set_pretrained_checkpoint(self, checkpoint: Dict[str, Any]) -> "FineTuningTask":
        assert (
            "classy_state_dict" in checkpoint
        ), "Checkpoint does not contain classy_state_dict"
        self.pretrained_checkpoint = checkpoint
        return self

    def set_reset_heads(self, reset_heads: bool) -> "FineTuningTask":
        self.reset_heads = reset_heads
        return self

    def set_freeze_trunk(self, freeze_trunk: bool) -> "FineTuningTask":
        self.freeze_trunk = freeze_trunk
        return self

    def _set_model_train_mode(self):
        phase = self.phases[self.phase_idx]
        if self.freeze_trunk:
            # convert all the sub-modules to the eval mode, except the heads
            self.base_model.eval()
            for heads in self.base_model.get_heads().values():
                for h in heads.values():
                    h.train(phase["train"])
        else:
            self.base_model.train(phase["train"])

    def prepare(
        self,
        num_dataloader_workers: int = 0,
        pin_memory: bool = False,
        use_gpu: bool = False,
        dataloader_mp_context=None,
    ) -> None:
        assert (
            self.pretrained_checkpoint is not None
        ), "Need a pretrained checkpoint for fine tuning"
        super().prepare(
            num_dataloader_workers, pin_memory, use_gpu, dataloader_mp_context
        )
        if self.checkpoint is None:
            # no checkpoint exists, load the model's state from the pretrained
            # checkpoint
            state_load_success = update_classy_model(
                self.base_model,
                self.pretrained_checkpoint["classy_state_dict"]["base_model"],
                self.reset_heads,
            )
            assert (
                state_load_success
            ), "Update classy state from pretrained checkpoint was unsuccessful."

        if self.freeze_trunk:
            # do not track gradients for all the parameters in the model except
            # for the parameters in the heads
            for param in self.base_model.parameters():
                param.requires_grad = False
            for heads in self.base_model.get_heads().values():
                for h in heads.values():
                    for param in h.parameters():
                        param.requires_grad = True
            # re-create ddp model
            self.distributed_model = None
            self.init_distributed_data_parallel_model()



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.


import copy
from pathlib import Path
from typing import Any, Callable, Dict, List

import torchvision.transforms as transforms
import torchvision.transforms._transforms_video as transforms_video
from classy_vision.generic.registry_utils import import_all_modules

from .classy_transform import ClassyTransform


FILE_ROOT = Path(__file__).parent


TRANSFORM_REGISTRY = {}


def build_transform(transform_config: Dict[str, Any]) -> Callable:
    """Builds a :class:`ClassyTransform` from a config.

    This assumes a 'name' key in the config which is used to determine what
    transform class to instantiate. For instance, a config `{"name":
    "my_transform", "foo": "bar"}` will find a class that was registered as
    "my_transform" (see :func:`register_transform`) and call .from_config on
    it.

    In addition to transforms registered with :func:`register_transform`, we
    also support instantiating transforms available in the
    `torchvision.transforms <https://pytorch.org/docs/stable/torchvision/
    transforms.html>`_ module. Any keys in the config will get expanded
    to parameters of the transform constructor. For instance, the following
    call will instantiate a :class:`torchvision.transforms.CenterCrop`:

    .. code-block:: python

      build_transform({"name": "CenterCrop", "size": 224})
    """
    assert (
        "name" in transform_config
    ), f"name not provided for transform: {transform_config}"
    name = transform_config["name"]
    transform_args = copy.deepcopy(transform_config)
    del transform_args["name"]
    if name in TRANSFORM_REGISTRY:
        return TRANSFORM_REGISTRY[name].from_config(transform_args)
    # the name should be available in torchvision.transforms
    assert hasattr(transforms, name) or hasattr(transforms_video, name), (
        f"{name} isn't a registered tranform"
        ", nor is it available in torchvision.transforms"
    )
    if hasattr(transforms, name):
        return getattr(transforms, name)(**transform_args)
    else:
        return getattr(transforms_video, name)(**transform_args)


def build_transforms(transforms_config: List[Dict[str, Any]]) -> Callable:
    """
    Builds a transform from the list of transform configurations.
    """
    transform_list = [build_transform(config) for config in transforms_config]
    return transforms.Compose(transform_list)


def register_transform(name: str):
    """Registers a :class:`ClassyTransform` subclass.

    This decorator allows Classy Vision to instantiate a subclass of
    :class:`ClassyTransform` from a configuration file, even if the class itself is not
    part of the Classy Vision framework. To use it, apply this decorator to a
    ClassyTransform subclass like this:

    .. code-block:: python

      @register_transform("my_transform")
      class MyTransform(ClassyTransform):
          ...

    To instantiate a transform from a configuration file, see
    :func:`build_transform`."""

    def register_transform_cls(cls: Callable[..., Callable]):
        if name in TRANSFORM_REGISTRY:
            raise ValueError("Cannot register duplicate transform ({})".format(name))
        TRANSFORM_REGISTRY[name] = cls
        return cls

    return register_transform_cls


# automatically import any Python files in the transforms/ directory
import_all_modules(FILE_ROOT, "classy_vision.dataset.transforms")

from .lighting_transform import LightingTransform  # isort:skip
from .util import ApplyTransformToKey  # isort:skip
from .util import ImagenetAugmentTransform  # isort:skip
from .util import ImagenetAugmentTransform  # isort:skip
from .util import ImagenetNoAugmentTransform  # isort:skip
from .util import GenericImageTransform  # isort:skip
from .util import TupleToMapTransform  # isort:skip


__all__ = [
    "ClassyTransform",
    "ImagenetAugmentTransform",
    "ImagenetNoAugmentTransform",
    "GenericImageTransform",
    "ApplyTransformToKey",
    "TupleToMapTransform",
    "LightingTransform",
    "register_transform",
    "build_transform",
    "build_transforms",
]



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import torch

from ...generic.util import torch_seed


class RandomVideoDataset:
    def __init__(
        self,
        num_classes,
        split,
        num_samples,
        frames_per_clip,
        video_width,
        video_height,
        audio_samples,
        clips_per_video,
        seed=10,
    ):
        self.num_classes = num_classes
        self.split = split
        # video config
        self.video_channels = 3
        self.num_samples = num_samples
        self.frames_per_clip = frames_per_clip
        self.video_width = video_width
        self.video_height = video_height
        # audio config
        self.audio_samples = audio_samples
        self.clips_per_video = clips_per_video
        # misc config
        self.seed = seed

    def __getitem__(self, idx):
        if self.split == "train":
            # assume we only sample 1 clip from each training video
            target_seed_offset = idx
        else:
            # for video model testing, clips from the same video share the same
            # target label
            target_seed_offset = idx // self.clips_per_video
        with torch_seed(self.seed + target_seed_offset):
            target = torch.randint(0, self.num_classes, (1,)).item()

        with torch_seed(self.seed + idx):
            return {
                "input": {
                    "video": torch.randint(
                        0,
                        256,
                        (
                            self.frames_per_clip,
                            self.video_height,
                            self.video_width,
                            self.video_channels,
                        ),
                        dtype=torch.uint8,
                    ),
                    "audio": torch.rand((self.audio_samples, 1), dtype=torch.float),
                },
                "target": target,
            }

    def __len__(self):
        return self.num_samples



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from enum import Enum, auto


class ClassyHookFunctions(Enum):
    """
    Enumeration of all the hook functions in the ClassyHook class.
    """

    on_start = auto()
    on_phase_start = auto()
    on_forward = auto()
    on_loss_and_meter = auto()
    on_step = auto()
    on_phase_end = auto()
    on_end = auto()



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import multiprocessing
import queue
import tempfile
import unittest

import torch


UPDATE_SIGNAL = 0
VALUE_SIGNAL = 1
SHUTDOWN_SIGNAL = 2
TIMEOUT = 100


def _get_value_or_raise_error(qout, qerr):
    try:
        return qout.get(timeout=TIMEOUT)
    except queue.Empty:
        raise qerr.get(timeout=TIMEOUT)


def _run(qin, qout, qerr, func, *args):
    try:
        func(qin, qout, *args)
    except Exception as e:
        print(e)
        qerr.put(e)


def _meter_worker(qin, qout, meter, is_train, world_size, rank, filename):
    backend = "gloo"
    torch.distributed.init_process_group(
        backend=backend,
        init_method="file://{filename}".format(filename=filename),
        world_size=world_size,
        rank=rank,
    )
    # Listen for commands on queues
    while True:
        try:
            signal, val = qin.get()
        except queue.Empty:
            continue

        if signal == UPDATE_SIGNAL:
            meter.update(val[0], val[1], is_train=is_train)

        elif signal == VALUE_SIGNAL:
            meter.sync_state()
            qout.put(meter.value)

        elif signal == SHUTDOWN_SIGNAL:
            break

        else:
            raise NotImplementedError("Bad signal value")

    return


class ClassificationMeterTest(unittest.TestCase):
    def setUp(self):
        self.mp = multiprocessing.get_context("spawn")
        self.processes = []

    def tearDown(self):
        for p in self.processes:
            p.terminate()

    def _spawn(self, func, *args):
        name = "process #%d" % len(self.processes)
        qin = self.mp.Queue()
        qout = self.mp.Queue()
        qerr = self.mp.Queue()
        qio = (qin, qout, qerr)
        args = qio + (func,) + args
        process = self.mp.Process(target=_run, name=name, args=args, daemon=True)
        process.start()
        self.processes.append(process)
        return qio

    def _apply_updates_and_test_meter(
        self, meter, model_output, target, expected_value, **kwargs
    ):
        """
        Runs a valid meter test. Does not reset meter before / after running
        """
        if not isinstance(model_output, list):
            model_output = [model_output]

        if not isinstance(target, list):
            target = [target]

        for i in range(len(model_output)):
            meter.update(model_output[i], target[i], **kwargs)

        meter.sync_state()
        meter_value = meter.value
        for key, val in expected_value.items():
            self.assertTrue(
                key in meter_value, msg="{0} not in meter value!".format(key)
            )
            if torch.is_tensor(meter_value[key]):
                self.assertTrue(
                    torch.all(torch.eq(meter_value[key], val)),
                    msg="{0} meter value mismatch!".format(key),
                )
            else:
                self.assertAlmostEqual(
                    meter_value[key],
                    val,
                    places=4,
                    msg="{0} meter value mismatch!".format(key),
                )

    def _values_match_expected_value(self, value0, value1, expected_value):
        for key, val in expected_value.items():
            self.assertTrue(key in value0, msg="{0} not in meter value!".format(key))
            self.assertTrue(key in value1, msg="{0} not in meter value!".format(key))
            if torch.is_tensor(val):
                self.assertTrue(
                    torch.all(torch.eq(value0[key], val)),
                    "{0} meter value mismatch!".format(key),
                )
                self.assertTrue(
                    torch.all(torch.eq(value1[key], val)),
                    "{0} meter value mismatch!".format(key),
                )
            else:
                self.assertAlmostEqual(
                    value0[key],
                    val,
                    places=4,
                    msg="{0} meter value mismatch!".format(key),
                )
                self.assertAlmostEqual(
                    value1[key],
                    val,
                    places=4,
                    msg="{0} meter value mismatch!".format(key),
                )

    def meter_update_and_reset_test(
        self, meter, model_outputs, targets, expected_value, **kwargs
    ):
        """
        This test verifies that a single update on the meter is successful,
        resets the meter, then applies the update again.
        """
        # If a single output is provided, wrap in list
        if not isinstance(model_outputs, list):
            model_outputs = [model_outputs]
            targets = [targets]

        for i in range(len(model_outputs)):
            meter.validate(model_outputs[i].size(), targets[i].size())

        self._apply_updates_and_test_meter(
            meter, model_outputs, targets, expected_value, **kwargs
        )

        meter.reset()

        # Verify reset works by reusing single update test
        self._apply_updates_and_test_meter(
            meter, model_outputs, targets, expected_value, **kwargs
        )

    def meter_invalid_meter_input_test(self, meter, model_output, target):
        # Invalid model
        with self.assertRaises(AssertionError):
            meter.validate(model_output.shape, target.shape)

    def meter_invalid_update_test(self, meter, model_output, target, **kwargs):
        """
        Runs a valid meter test. Does not reset meter before / after running
        """
        if not isinstance(model_output, list):
            model_output = [model_output]

        if not isinstance(target, list):
            target = [target]

        with self.assertRaises(AssertionError):
            for i in range(len(model_output)):
                meter.update(model_output[i], target[i], **kwargs)

    def meter_get_set_classy_state_test(
        self, meters, model_outputs, targets, expected_value, **kwargs
    ):
        """
        Tests get and set classy state methods of meter.
        """
        assert len(meters) == 2, "Incorrect number of meters passed to test"
        assert (
            len(model_outputs) == 2
        ), "Incorrect number of model_outputs passed to test"
        assert len(targets) == 2, "Incorrect number of targets passed to test"
        meter0 = meters[0]
        meter1 = meters[1]

        meter0.update(model_outputs[0], targets[0], **kwargs)
        meter1.update(model_outputs[1], targets[1], **kwargs)

        meter0.sync_state()
        value0 = meter0.value

        meter1.sync_state()
        value1 = meter1.value
        for key, val in value0.items():
            if torch.is_tensor(value1[key]):
                self.assertFalse(
                    torch.all(torch.eq(value1[key], val)),
                    msg="{0} meter values should not be same!".format(key),
                )
            else:
                self.assertNotEqual(
                    value1[key],
                    val,
                    msg="{0} meter values should not be same!".format(key),
                )

        meter0.set_classy_state(meter1.get_classy_state())
        value0 = meter0.value
        for key, val in value0.items():
            if torch.is_tensor(value1[key]):
                self.assertTrue(
                    torch.all(torch.eq(value1[key], val)),
                    msg="{0} meter value mismatch after state transfer!".format(key),
                )
                self.assertTrue(
                    torch.all(torch.eq(value1[key], expected_value[key])),
                    msg="{0} meter value mismatch from ground truth!".format(key),
                )
            else:
                self.assertAlmostEqual(
                    value1[key],
                    val,
                    places=4,
                    msg="{0} meter value mismatch after state transfer!".format(key),
                )
                self.assertAlmostEqual(
                    value1[key],
                    expected_value[key],
                    places=4,
                    msg="{0} meter value mismatch from ground truth!".format(key),
                )

    def _spawn_all_meter_workers(self, world_size, meters, is_train):
        filename = tempfile.NamedTemporaryFile(delete=True).name
        qins = []
        qerrs = []
        qouts = []

        for i in range(world_size):
            qin, qout, qerr = self._spawn(
                _meter_worker, meters[i], is_train, world_size, i, filename
            )
            qins.append(qin)
            qouts.append(qout)
            qerrs.append(qerr)

        return qins, qouts, qerrs

    def meter_distributed_test(
        self, meters, model_outputs, targets, expected_values, is_train=False
    ):
        """
        Sets up two processes each with a given meter on that process.
        Verifies that sync code path works.
        """
        world_size = len(meters)
        assert world_size == 2, "This test only works for world_size of 2"
        assert len(model_outputs) == 4, (
            "Test assumes 4 model outputs, "
            "0, 2 passed to meter0 and 1, 3 passed to meter1"
        )
        assert (
            len(targets) == 4
        ), "Test assumes 4 targets, 0, 2 passed to meter0 and 1, 3 passed to meter1"
        assert len(expected_values) == 2, (
            "Test assumes 2 expected values, "
            "first is result of applying updates 0,1 to the meter, "
            "second is result of applying all 4 updates to meter"
        )

        qins, qouts, qerrs = self._spawn_all_meter_workers(
            world_size, meters, is_train=is_train
        )

        # First update each meter, then get value from each meter
        qins[0].put_nowait((UPDATE_SIGNAL, (model_outputs[0], targets[0])))
        qins[1].put_nowait((UPDATE_SIGNAL, (model_outputs[1], targets[1])))

        qins[0].put_nowait((VALUE_SIGNAL, None))
        qins[1].put_nowait((VALUE_SIGNAL, None))

        value0 = _get_value_or_raise_error(qouts[0], qerrs[0])
        value1 = _get_value_or_raise_error(qouts[1], qerrs[1])
        self._values_match_expected_value(value0, value1, expected_values[0])

        # Verify that calling value again does not break things
        qins[0].put_nowait((VALUE_SIGNAL, None))
        qins[1].put_nowait((VALUE_SIGNAL, None))

        value0 = _get_value_or_raise_error(qouts[0], qerrs[0])
        value1 = _get_value_or_raise_error(qouts[1], qerrs[1])
        self._values_match_expected_value(value0, value1, expected_values[0])

        # Second, update each meter, then get value from each meter
        qins[0].put_nowait((UPDATE_SIGNAL, (model_outputs[2], targets[2])))
        qins[1].put_nowait((UPDATE_SIGNAL, (model_outputs[3], targets[3])))

        qins[0].put_nowait((VALUE_SIGNAL, None))
        qins[1].put_nowait((VALUE_SIGNAL, None))

        value0 = _get_value_or_raise_error(qouts[0], qerrs[0])
        value1 = _get_value_or_raise_error(qouts[1], qerrs[1])
        self._values_match_expected_value(value0, value1, expected_values[1])

        qins[0].put_nowait((SHUTDOWN_SIGNAL, None))
        qins[1].put_nowait((SHUTDOWN_SIGNAL, None))



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import shutil
import tempfile
import unittest
from test.generic.config_utils import get_test_task_config

import torch
from classy_vision.dataset.transforms import ClassyTransform
from classy_vision.hub import ClassyHubInterface
from classy_vision.models import ClassyModel, build_model
from classy_vision.tasks import ClassyTask, build_task
from torchvision import models, transforms


class TestTransform(ClassyTransform):
    def __call__(self, x):
        return x


class TestClassyHubInterface(unittest.TestCase):
    def setUp(self):
        # create a base directory to write image files to
        self.base_dir = tempfile.mkdtemp()
        self.image_path = self.base_dir + "/img.jpg"
        # create an image with a non standard size
        image_tensor = torch.zeros((3, 1000, 2500), dtype=torch.float)
        transforms.ToPILImage()(image_tensor).save(self.image_path)

    def tearDown(self):
        # delete all the temporary data created
        shutil.rmtree(self.base_dir)

    def _test_predict_and_extract_features(self, hub_interface: ClassyHubInterface):
        dataset = hub_interface.create_image_dataset(
            [self.image_path], phase_type="test"
        )
        data_iterator = hub_interface.get_data_iterator(dataset)
        input = next(data_iterator)
        # set the model to eval mode
        hub_interface.eval()
        output = hub_interface.predict(input)
        self.assertIsNotNone(output)
        # see the prediction for the input
        hub_interface.predict(input).argmax().item()
        # check extract features
        output = hub_interface.extract_features(input)
        self.assertIsNotNone(output)

    def _get_classy_model(self):
        config = get_test_task_config()
        model_config = config["model"]
        return build_model(model_config)

    def _get_non_classy_model(self):
        return models.resnet18(pretrained=False)

    def test_from_task(self):
        config = get_test_task_config()
        task = build_task(config)
        hub_interface = ClassyHubInterface.from_task(task)

        self.assertIsInstance(hub_interface.task, ClassyTask)
        self.assertIsInstance(hub_interface.model, ClassyModel)

        # this will pick up the transform from the task's config
        self._test_predict_and_extract_features(hub_interface)

        # test that the correct transform is picked up
        phase_type = "test"
        test_transform = TestTransform()
        task.datasets[phase_type].transform = test_transform
        hub_interface = ClassyHubInterface.from_task(task)
        dataset = hub_interface.create_image_dataset(
            [self.image_path], phase_type=phase_type
        )
        self.assertIsInstance(dataset.transform, TestTransform)

    def test_from_model(self):
        for model in [self._get_classy_model(), self._get_non_classy_model()]:
            hub_interface = ClassyHubInterface.from_model(model)

            self.assertIsNone(hub_interface.task)
            self.assertIsInstance(hub_interface.model, ClassyModel)

            # this will pick up the transform from imagenet
            self._test_predict_and_extract_features(hub_interface)



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import collections.abc as abc
import logging
import operator

import torch
import torch.nn as nn
from classy_vision.generic.util import get_model_dummy_input, is_leaf, is_on_gpu
from torch.cuda import cudart


def profile(
    model,
    batchsize_per_replica=32,
    input_shape=(3, 224, 224),
    use_nvprof=False,
    input_key=None,
):
    """
    Performs CPU or GPU profiling of the specified model on the specified input.
    """

    # assertions:
    if use_nvprof:
        raise NotImplementedError
        # FIXME (mannatsingh): in case of use_nvprof, exit() is called at the end
        # and we do not return a profile.
        assert is_on_gpu(model), "can only nvprof model that lives on GPU"
        logging.info("CUDA profiling: Make sure you are running under nvprof!")

    # input for model:
    input = get_model_dummy_input(
        model,
        input_shape,
        input_key,
        batchsize=batchsize_per_replica,
        non_blocking=False,
    )
    # perform profiling:
    with torch.no_grad():
        model(input)  # warm up CUDA memory allocator and profiler
        if use_nvprof:  # nvprof profiling (TODO: Can we infer this?)
            cudart().cudaProfilerStart()
            model(input)
            cudart().cudaProfilerStop()
            exit()  # exit gracefully
        else:  # regular profiling
            with torch.autograd.profiler.profile(use_cuda=True) as profiler:
                model(input)
                return profiler


def _layer_flops(layer, x, _):
    """
    Computes the number of FLOPs required for a single layer.

    For common layers, such as Conv1d, the flop compute is implemented in this
    centralized place.
    For other layers, if it defines a method to compute flops with the signature
    below, we will use it to compute flops.

    Class MyModule(nn.Module):
        def flops(self, x):
            ...

    """

    # get layer type:
    typestr = layer.__repr__()
    layer_type = typestr[: typestr.find("(")].strip()
    batchsize_per_replica = x.size()[0]
    # 1D convolution:
    if layer_type in ["Conv1d"]:
        # x shape is N x C x W
        out_w = int(
            (x.size()[2] + 2 * layer.padding[0] - layer.kernel_size[0])
            / layer.stride[0]
            + 1
        )
        return (
            batchsize_per_replica
            * layer.in_channels
            * layer.out_channels
            * layer.kernel_size[0]
            * out_w
            / layer.groups
        )
    # 2D convolution:
    elif layer_type in ["Conv2d"]:
        out_h = int(
            (x.size()[2] + 2 * layer.padding[0] - layer.kernel_size[0])
            / layer.stride[0]
            + 1
        )
        out_w = int(
            (x.size()[3] + 2 * layer.padding[1] - layer.kernel_size[1])
            / layer.stride[1]
            + 1
        )
        return (
            batchsize_per_replica
            * layer.in_channels
            * layer.out_channels
            * layer.kernel_size[0]
            * layer.kernel_size[1]
            * out_h
            * out_w
            / layer.groups
        )

    # learned group convolution:
    elif layer_type in ["LearnedGroupConv"]:
        conv = layer.conv
        out_h = int(
            (x.size()[2] + 2 * conv.padding[0] - conv.kernel_size[0]) / conv.stride[0]
            + 1
        )
        out_w = int(
            (x.size()[3] + 2 * conv.padding[1] - conv.kernel_size[1]) / conv.stride[1]
            + 1
        )
        count1 = _layer_flops(layer.relu, x) + _layer_flops(layer.norm, x)
        count2 = (
            batchsize_per_replica
            * conv.in_channels
            * conv.out_channels
            * conv.kernel_size[0]
            * conv.kernel_size[1]
            * out_h
            * out_w
            / layer.condense_factor
        )
        return count1 + count2

    # non-linearities:
    elif layer_type in ["ReLU", "ReLU6", "Tanh", "Sigmoid", "Softmax"]:
        return x.numel()

    # 2D pooling layers:
    elif layer_type in ["AvgPool2d", "MaxPool2d"]:
        in_h = x.size()[2]
        in_w = x.size()[3]
        if isinstance(layer.kernel_size, int):
            layer.kernel_size = (layer.kernel_size, layer.kernel_size)
        kernel_ops = layer.kernel_size[0] * layer.kernel_size[1]
        out_h = 1 + int(
            (in_h + 2 * layer.padding - layer.kernel_size[0]) / layer.stride
        )
        out_w = 1 + int(
            (in_w + 2 * layer.padding - layer.kernel_size[1]) / layer.stride
        )
        return x.size()[0] * x.size()[1] * out_w * out_h * kernel_ops

    # adaptive avg pool2d
    # This is approximate and works only for downsampling without padding
    # based on aten/src/ATen/native/AdaptiveAveragePooling.cpp
    elif layer_type in ["AdaptiveAvgPool2d"]:
        in_h = x.size()[2]
        in_w = x.size()[3]
        out_h = layer.output_size[0]
        out_w = layer.output_size[1]
        if out_h > in_h or out_w > in_w:
            raise NotImplementedError()
        batchsize_per_replica = x.size()[0]
        num_channels = x.size()[1]
        kh = in_h - out_h + 1
        kw = in_w - out_w + 1
        kernel_ops = kh * kw
        return batchsize_per_replica * num_channels * out_h * out_w * kernel_ops

    # linear layer:
    elif layer_type in ["Linear"]:
        weight_ops = layer.weight.numel()
        bias_ops = layer.bias.numel() if layer.bias is not None else 0
        return x.size()[0] * (weight_ops + bias_ops)

    # 2D/3D batch normalization:
    elif layer_type in ["BatchNorm2d", "BatchNorm3d"]:
        return 2 * x.numel()

    # 3D convolution
    elif layer_type in ["Conv3d"]:
        out_t = int(
            (x.size()[2] + 2 * layer.padding[0] - layer.kernel_size[0])
            // layer.stride[0]
            + 1
        )
        out_h = int(
            (x.size()[3] + 2 * layer.padding[1] - layer.kernel_size[1])
            // layer.stride[1]
            + 1
        )
        out_w = int(
            (x.size()[4] + 2 * layer.padding[2] - layer.kernel_size[2])
            // layer.stride[2]
            + 1
        )
        return (
            batchsize_per_replica
            * layer.in_channels
            * layer.out_channels
            * layer.kernel_size[0]
            * layer.kernel_size[1]
            * layer.kernel_size[2]
            * out_t
            * out_h
            * out_w
            / layer.groups
        )

    # 3D pooling layers
    elif layer_type in ["AvgPool3d", "MaxPool3d"]:
        in_t = x.size()[2]
        in_h = x.size()[3]
        in_w = x.size()[4]
        if isinstance(layer.kernel_size, int):
            layer.kernel_size = (
                layer.kernel_size,
                layer.kernel_size,
                layer.kernel_size,
            )
        if isinstance(layer.padding, int):
            layer.padding = (layer.padding, layer.padding, layer.padding)
        if isinstance(layer.stride, int):
            layer.stride = (layer.stride, layer.stride, layer.stride)
        kernel_ops = layer.kernel_size[0] * layer.kernel_size[1] * layer.kernel_size[2]
        out_t = 1 + int(
            (in_t + 2 * layer.padding[0] - layer.kernel_size[0]) / layer.stride[0]
        )
        out_h = 1 + int(
            (in_h + 2 * layer.padding[1] - layer.kernel_size[1]) / layer.stride[1]
        )
        out_w = 1 + int(
            (in_w + 2 * layer.padding[2] - layer.kernel_size[2]) / layer.stride[2]
        )
        return batchsize_per_replica * x.size()[1] * out_t * out_h * out_w * kernel_ops

    # adaptive avg pool3d
    # This is approximate and works only for downsampling without padding
    # based on aten/src/ATen/native/AdaptiveAveragePooling3d.cpp
    elif layer_type in ["AdaptiveAvgPool3d"]:
        in_t = x.size()[2]
        in_h = x.size()[3]
        in_w = x.size()[4]
        out_t = layer.output_size[0]
        out_h = layer.output_size[1]
        out_w = layer.output_size[2]
        if out_t > in_t or out_h > in_h or out_w > in_w:
            raise NotImplementedError()
        batchsize_per_replica = x.size()[0]
        num_channels = x.size()[1]
        kt = in_t - out_t + 1
        kh = in_h - out_h + 1
        kw = in_w - out_w + 1
        kernel_ops = kt * kh * kw
        return batchsize_per_replica * num_channels * out_t * out_w * out_h * kernel_ops

    # dropout layer
    elif layer_type in ["Dropout"]:
        # At test time, we do not drop values but scale the feature map by the
        # dropout ratio
        flops = 1
        for dim_size in x.size():
            flops *= dim_size
        return flops
    elif hasattr(layer, "flops"):
        # If the module already defines a method to compute flops with the signature
        # below, we use it to compute flops
        #
        #   Class MyModule(nn.Module):
        #     def flops(self, x):
        #       ...
        return layer.flops(x)

    # not implemented:
    raise NotImplementedError("FLOPs not implemented for %s layer" % layer_type)


def _layer_activations(layer, x, out):
    """
    Computes the number of activations produced by a single layer.

    Activations are counted only for convolutional layers.
    """
    return out.numel() if isinstance(layer, (nn.Conv1d, nn.Conv2d, nn.Conv3d)) else 0


def summarize_profiler_info(prof):
    """
    Summarizes the statistics in the specified profiler.
    """

    # create sorted list of times per operator:
    op2time = {}
    for item in prof.key_averages():
        op2time[item.key] = (
            item.cpu_time_total / 1000.0,
            item.cuda_time_total / 1000.0,
        )  # to milliseconds
    op2time = sorted(op2time.items(), key=operator.itemgetter(1), reverse=True)

    # created string containing information:
    str = "\n%s\tCPU Time\tCUDA Time\n" % ("Key".rjust(20))
    for (key, value) in op2time:
        str += "%s\t%2.5f ms\t%2.5f ms\n" % (key.rjust(20), value[0], value[1])
    return str


def _patched_computation_module(module, compute_list, compute_fn):
    """
    Patch the module to compute a module's parameters, like FLOPs.

    Calls compute_fn and appends the results to compute_list.
    """
    ty = type(module)
    typestring = module.__repr__()

    class ComputeModule(ty):
        orig_type = ty

        def _original_forward(self, *args, **kwargs):
            return ty.forward(self, *args, **kwargs)

        def forward(self, *args, **kwargs):
            out = self._original_forward(*args, **kwargs)
            compute_list.append(compute_fn(self, args[0], out))
            return out

        def __repr__(self):
            return typestring

    return ComputeModule


def modify_forward(model, compute_list, compute_fn):
    """
    Modify forward pass to measure a module's parameters, like FLOPs.
    """
    if is_leaf(model):
        model.__class__ = _patched_computation_module(model, compute_list, compute_fn)
    for child in model.children():
        modify_forward(child, compute_list, compute_fn)

    return model


def restore_forward(model):
    """
    Restore original forward in model:
    """
    if is_leaf(model):
        model.__class__ = model.orig_type
    for child in model.children():
        restore_forward(child)

    return model


def compute_complexity(model, compute_fn, input_shape, input_key=None):
    """
    Compute the complexity of a forward pass.
    """

    # assertions, input, and upvalue in which we will perform the count:
    assert isinstance(model, nn.Module)
    if not isinstance(input_shape, abc.Sequence):
        return None
    input = get_model_dummy_input(model, input_shape, input_key)
    compute_list = []

    # measure FLOPs:
    modify_forward(model, compute_list, compute_fn)
    try:
        model.forward(input)
    except NotImplementedError as err:
        raise err
    finally:
        restore_forward(model)
    return sum(compute_list)


def compute_flops(model, input_shape=(3, 224, 224), input_key=None):
    """
    Compute the number of FLOPs needed for a forward pass.
    """
    return compute_complexity(model, _layer_flops, input_shape, input_key)


def compute_activations(model, input_shape=(3, 224, 224), input_key=None):
    """
    Compute the number of activations created in a forward pass.
    """
    return compute_complexity(model, _layer_activations, input_shape, input_key)


def count_params(model):
    """
    Count the number of parameters in a model.
    """
    assert isinstance(model, nn.Module)
    count = 0
    for child in model.children():
        if is_leaf(child):
            if hasattr(child, "_mask"):  # for masked modules (like LGC)
                count += child._mask.long().sum().item()
                # FIXME: BatchNorm parameters in LGC are not counted.
            else:  # for regular modules
                for p in child.parameters():
                    count += p.nelement()
        else:
            count += count_params(child)
    return count



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import unittest
from test.generic.utils import compare_model_state

import torch
from classy_vision.models import build_model


MODELS = {
    "small_densenet": {
        "name": "densenet",
        "num_blocks": [1, 1, 1, 1],
        "init_planes": 4,
        "growth_rate": 32,
        "expansion": 4,
        "final_bn_relu": True,
        "small_input": True,
        "heads": [
            {
                "name": "fully_connected",
                "unique_id": "default_head",
                "num_classes": 1000,
                "fork_block": "trunk_output",
                "in_plane": 60,
                "zero_init_bias": True,
            }
        ],
    }
}


class TestDensenet(unittest.TestCase):
    def _test_model(self, model_config):
        """This test will build Densenet models, run a forward pass and
        verify output shape, and then verify that get / set state
        works.

        I do this in one test so that we construct the model a minimum
        number of times.
        """
        model = build_model(model_config)

        # Verify forward pass works
        input = torch.ones([1, 3, 32, 32])
        output = model.forward(input)
        self.assertEqual(output.size(), (1, 1000))

        # Verify get_set_state
        new_model = build_model(model_config)
        state = model.get_classy_state()
        new_model.set_classy_state(state)
        new_state = new_model.get_classy_state()

        compare_model_state(self, state, new_state, check_heads=True)

    def test_small_densenet(self):
        self._test_model(MODELS["small_densenet"])



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import copy
import unittest
from test.generic.config_utils import get_fast_test_task_config
from test.generic.utils import compare_model_state
from unittest import mock

from classy_vision.generic.util import get_checkpoint_dict
from classy_vision.tasks import FineTuningTask, build_task
from classy_vision.trainer import LocalTrainer


class TestFineTuningTask(unittest.TestCase):
    def _compare_model_state(self, state_1, state_2, check_heads=True):
        return compare_model_state(self, state_1, state_2, check_heads=check_heads)

    def _get_fine_tuning_config(
        self, head_num_classes=1000, pretrained_checkpoint=False
    ):
        config = get_fast_test_task_config(head_num_classes=head_num_classes)
        config["name"] = "fine_tuning"
        config["num_epochs"] = 10

        if pretrained_checkpoint:
            config["pretrained_checkpoint"] = "/path/to/pretrained/checkpoint"

        return config

    def _get_pre_train_config(self, head_num_classes=1000):
        config = get_fast_test_task_config(head_num_classes=head_num_classes)
        config["num_epochs"] = 10
        return config

    def test_build_task(self):
        config = self._get_fine_tuning_config()
        task = build_task(config)
        self.assertIsInstance(task, FineTuningTask)

        config = self._get_fine_tuning_config(pretrained_checkpoint=True)

        with mock.patch("classy_vision.tasks.FineTuningTask.set_pretrained_checkpoint"):
            task = build_task(config)

        self.assertIsInstance(task, FineTuningTask)

    def test_prepare(self):
        pre_train_config = self._get_pre_train_config()
        pre_train_task = build_task(pre_train_config)
        pre_train_task.prepare()
        checkpoint = get_checkpoint_dict(pre_train_task, {})

        fine_tuning_config = self._get_fine_tuning_config()
        fine_tuning_task = build_task(fine_tuning_config)

        # test: cannot prepare a fine tuning task without a pre-trained checkpoint
        with self.assertRaises(Exception):
            fine_tuning_task.prepare()

        # test: prepare should succeed after pre-trained checkpoint is set
        fine_tuning_task.set_pretrained_checkpoint(checkpoint)
        fine_tuning_task.prepare()

        # test: prepare should succeed if a pre-trained checkpoint is provided in the
        # config
        fine_tuning_config = self._get_fine_tuning_config(pretrained_checkpoint=True)
        with mock.patch(
            "classy_vision.tasks.fine_tuning_task.load_checkpoint",
            return_value=checkpoint,
        ):
            fine_tuning_task = build_task(fine_tuning_config)

        fine_tuning_task.prepare()

        # test: a fine tuning task with incompatible heads with a manually set
        # pre-trained checkpoint should fail to prepare if the heads are not reset
        fine_tuning_config = self._get_fine_tuning_config(head_num_classes=10)
        fine_tuning_task = build_task(fine_tuning_config)
        fine_tuning_task.set_pretrained_checkpoint(checkpoint)

        with self.assertRaises(Exception):
            fine_tuning_task.prepare()

        # test: a fine tuning task with incompatible heads with a manually set
        # pre-trained checkpoint should succeed to prepare if the heads are reset
        fine_tuning_task.set_pretrained_checkpoint(
            copy.deepcopy(checkpoint)
        ).set_reset_heads(True)

        fine_tuning_task.prepare()

        # test: a fine tuning task with incompatible heads with the pre-trained
        # checkpoint provided in the config should fail to prepare
        fine_tuning_config = self._get_fine_tuning_config(
            head_num_classes=10, pretrained_checkpoint=True
        )

        with mock.patch(
            "classy_vision.tasks.fine_tuning_task.load_checkpoint",
            return_value=copy.deepcopy(checkpoint),
        ):
            fine_tuning_task = build_task(fine_tuning_config)

        with self.assertRaises(Exception):
            fine_tuning_task.prepare()

        # test: a fine tuning task with incompatible heads with the pre-trained
        # checkpoint provided in the config should succeed to prepare if the heads are
        # reset
        fine_tuning_task.set_reset_heads(True)
        fine_tuning_task.prepare()

    def test_train(self):
        pre_train_config = self._get_pre_train_config(head_num_classes=1000)
        pre_train_task = build_task(pre_train_config)
        trainer = LocalTrainer()
        trainer.train(pre_train_task)
        checkpoint = get_checkpoint_dict(pre_train_task, {})

        for reset_heads, heads_num_classes in [(False, 1000), (True, 200)]:
            for freeze_trunk in [True, False]:
                fine_tuning_config = self._get_fine_tuning_config(
                    head_num_classes=heads_num_classes
                )
                fine_tuning_task = build_task(fine_tuning_config)
                fine_tuning_task = (
                    fine_tuning_task.set_pretrained_checkpoint(
                        copy.deepcopy(checkpoint)
                    )
                    .set_reset_heads(reset_heads)
                    .set_freeze_trunk(freeze_trunk)
                )
                # run in test mode to compare the model state
                fine_tuning_task.set_test_only(True)
                trainer.train(fine_tuning_task)
                self._compare_model_state(
                    pre_train_task.model.get_classy_state(),
                    fine_tuning_task.model.get_classy_state(),
                    check_heads=not reset_heads,
                )
                # run in train mode to check accuracy
                fine_tuning_task.set_test_only(False)
                trainer.train(fine_tuning_task)
                if freeze_trunk:
                    # if trunk is frozen the states should be the same
                    self._compare_model_state(
                        pre_train_task.model.get_classy_state(),
                        fine_tuning_task.model.get_classy_state(),
                        check_heads=False,
                    )
                else:
                    # trunk isn't frozen, the states should be different
                    with self.assertRaises(Exception):
                        self._compare_model_state(
                            pre_train_task.model.get_classy_state(),
                            fine_tuning_task.model.get_classy_state(),
                            check_heads=False,
                        )

                accuracy = fine_tuning_task.meters[0].value["top_1"]
                self.assertAlmostEqual(accuracy, 1.0)



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from typing import Any, Dict

import torch.optim
from classy_vision.generic.util import is_pos_float
from torch.optim import Optimizer

from . import ClassyOptimizer, register_optimizer


class RMSpropTFOptimizer(Optimizer):
    r"""Implements RMSprop algorithm.

    NOTE: This code is copied from :class:`torch.optim.RMSProp`, with the epsilon
    moved inside the square root to match tensorflow's implementation.

    Proposed by G. Hinton in his
    `course <http://www.cs.toronto.edu/~tijmen/csc321/slides/lecture_slides_lec6.pdf>`_.

    The centered version first appears in `Generating Sequences
    With Recurrent Neural Networks <https://arxiv.org/pdf/1308.0850v5.pdf>`_.

    The implementation here takes the square root of the gradient average before
    adding epsilon (note that TensorFlow interchanges these two operations). The
    effective learning rate is thus :math:`\alpha/(\sqrt{v} + \epsilon)` where
    :math:`\alpha` is the scheduled learning rate and :math:`v` is the weighted moving
    average of the squared gradient.

    Arguments:
        params (iterable): iterable of parameters to optimize or dicts defining
            parameter groups
        lr (float, optional): learning rate (default: 1e-2)
        momentum (float, optional): momentum factor (default: 0)
        alpha (float, optional): smoothing constant (default: 0.99)
        eps (float, optional): term added to the square root in the denominator to
            improve numerical stability (default: 1e-8)
        centered (bool, optional) : if ``True``, compute the centered RMSProp,
            the gradient is normalized by an estimation of its variance
        weight_decay (float, optional): weight decay (L2 penalty) (default: 0)

    """

    def __init__(
        self,
        params,
        lr=1e-2,
        alpha=0.99,
        eps=1e-8,
        weight_decay=0,
        momentum=0,
        centered=False,
    ):
        if not 0.0 <= lr:
            raise ValueError("Invalid learning rate: {}".format(lr))
        if not 0.0 <= eps:
            raise ValueError("Invalid epsilon value: {}".format(eps))
        if not 0.0 <= momentum:
            raise ValueError("Invalid momentum value: {}".format(momentum))
        if not 0.0 <= weight_decay:
            raise ValueError("Invalid weight_decay value: {}".format(weight_decay))
        if not 0.0 <= alpha:
            raise ValueError("Invalid alpha value: {}".format(alpha))

        defaults = dict(
            lr=lr,
            momentum=momentum,
            alpha=alpha,
            eps=eps,
            centered=centered,
            weight_decay=weight_decay,
        )
        super().__init__(params, defaults)

    def __setstate__(self, state):
        super().__setstate__(state)
        for group in self.param_groups:
            group.setdefault("momentum", 0)
            group.setdefault("centered", False)

    def step(self, closure=None):
        """Performs a single optimization step.

        Arguments:
            closure (callable, optional): A closure that reevaluates the model
                and returns the loss.
        """
        loss = None
        if closure is not None:
            loss = closure()

        for group in self.param_groups:
            for p in group["params"]:
                if p.grad is None:
                    continue
                grad = p.grad.data
                if grad.is_sparse:
                    raise RuntimeError("RMSprop does not support sparse gradients")
                state = self.state[p]

                # State initialization
                if len(state) == 0:
                    state["step"] = 0
                    state["square_avg"] = torch.zeros_like(
                        p.data, memory_format=torch.preserve_format
                    )
                    if group["momentum"] > 0:
                        state["momentum_buffer"] = torch.zeros_like(
                            p.data, memory_format=torch.preserve_format
                        )
                    if group["centered"]:
                        state["grad_avg"] = torch.zeros_like(
                            p.data, memory_format=torch.preserve_format
                        )

                square_avg = state["square_avg"]
                alpha = group["alpha"]

                state["step"] += 1

                if group["weight_decay"] != 0:
                    grad = grad.add(group["weight_decay"], p.data)

                square_avg.mul_(alpha).addcmul_(1 - alpha, grad, grad)

                if group["centered"]:
                    grad_avg = state["grad_avg"]
                    grad_avg.mul_(alpha).add_(1 - alpha, grad)
                    avg = (
                        square_avg.addcmul(-1, grad_avg, grad_avg)
                        .add_(group["eps"])
                        .sqrt_()
                    )
                else:
                    avg = square_avg.add_(group["eps"]).sqrt_()

                if group["momentum"] > 0:
                    buf = state["momentum_buffer"]
                    buf.mul_(group["momentum"]).addcdiv_(grad, avg)
                    p.data.add_(-group["lr"], buf)
                else:
                    p.data.addcdiv_(-group["lr"], grad, avg)

        return loss


@register_optimizer("rmsprop_tf")
class RMSPropTF(ClassyOptimizer):
    def __init__(
        self,
        lr: float = 0.1,
        momentum: float = 0,
        weight_decay: float = 0,
        alpha: float = 0.99,
        eps: float = 1e-8,
        centered: bool = False,
    ) -> None:
        super().__init__()

        self.parameters.lr = lr
        self.parameters.momentum = momentum
        self.parameters.weight_decay = weight_decay
        self.parameters.alpha = alpha
        self.parameters.eps = eps
        self.parameters.centered = centered

    def init_pytorch_optimizer(self, model, **kwargs):
        super().init_pytorch_optimizer(model, **kwargs)
        self.optimizer = RMSpropTFOptimizer(
            self.param_groups_override,
            lr=self.parameters.lr,
            momentum=self.parameters.momentum,
            weight_decay=self.parameters.weight_decay,
            alpha=self.parameters.alpha,
            eps=self.parameters.eps,
            centered=self.parameters.centered,
        )

    @classmethod
    def from_config(cls, config: Dict[str, Any]) -> "RMSPropTF":
        """Instantiates a RMSPropTF from a configuration.

        Args:
            config: A configuration for a RMSPropTF.
                See :func:`__init__` for parameters expected in the config.

        Returns:
            A RMSPropTF instance.
        """
        # Default params
        config.setdefault("lr", 0.1)
        config.setdefault("momentum", 0.0)
        config.setdefault("weight_decay", 0.0)
        config.setdefault("alpha", 0.99)
        config.setdefault("eps", 1e-8)
        config.setdefault("centered", False)

        for key in ["momentum", "alpha"]:
            assert (
                config[key] >= 0.0 and config[key] < 1.0 and type(config[key]) == float
            ), f"Config must contain a '{key}' in [0, 1) for RMSPropTF optimizer"
        assert is_pos_float(
            config["eps"]
        ), f"Config must contain a positive 'eps' for RMSPropTF optimizer"
        assert isinstance(
            config["centered"], bool
        ), "Config must contain a boolean 'centered' param for RMSPropTF optimizer"

        return cls(
            lr=config["lr"],
            momentum=config["momentum"],
            weight_decay=config["weight_decay"],
            alpha=config["alpha"],
            eps=config["eps"],
            centered=config["centered"],
        )



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import shutil
import tempfile
import unittest
from test.generic.config_utils import get_fast_test_task_config, get_test_task_config
from test.generic.utils import compare_model_state, compare_samples, compare_states

import torch
from classy_vision.dataset import build_dataset
from classy_vision.generic.util import get_checkpoint_dict, load_checkpoint
from classy_vision.hooks import CheckpointHook, LossLrMeterLoggingHook
from classy_vision.losses import build_loss
from classy_vision.models import build_model
from classy_vision.optim import build_optimizer
from classy_vision.tasks import ClassificationTask, build_task
from classy_vision.trainer import LocalTrainer


class TestClassificationTask(unittest.TestCase):
    def _compare_model_state(self, model_state_1, model_state_2, check_heads=True):
        compare_model_state(self, model_state_1, model_state_2, check_heads)

    def _compare_samples(self, sample_1, sample_2):
        compare_samples(self, sample_1, sample_2)

    def _compare_states(self, state_1, state_2, check_heads=True):
        compare_states(self, state_1, state_2)

    def setUp(self):
        # create a base directory to write checkpoints to
        self.base_dir = tempfile.mkdtemp()

    def tearDown(self):
        # delete all the temporary data created
        shutil.rmtree(self.base_dir)

    def test_build_task(self):
        config = get_test_task_config()
        task = build_task(config)
        self.assertTrue(isinstance(task, ClassificationTask))

    def test_get_state(self):
        config = get_test_task_config()
        loss = build_loss(config["loss"])
        task = (
            ClassificationTask()
            .set_num_epochs(1)
            .set_loss(loss)
            .set_model(build_model(config["model"]))
            .set_optimizer(build_optimizer(config["optimizer"]))
        )
        for phase_type in ["train", "test"]:
            dataset = build_dataset(config["dataset"][phase_type])
            task.set_dataset(dataset, phase_type)

        task.prepare(num_dataloader_workers=1, pin_memory=False)

        task = build_task(config)
        task.prepare(num_dataloader_workers=1, pin_memory=False)

    def test_checkpointing(self):
        """
        Tests checkpointing by running train_steps to make sure the train_steps
        run the same way after loading from a checkpoint.
        """
        config = get_fast_test_task_config()
        task = build_task(config).set_hooks([LossLrMeterLoggingHook()])
        task_2 = build_task(config).set_hooks([LossLrMeterLoggingHook()])

        use_gpu = torch.cuda.is_available()
        local_variables = {}

        # prepare the tasks for the right device
        task.prepare(use_gpu=use_gpu)

        # test in both train and test mode
        for _ in range(2):
            task.advance_phase()

            # set task's state as task_2's checkpoint
            task_2.set_checkpoint(get_checkpoint_dict(task, {}, deep_copy=True))
            task_2.prepare(use_gpu=use_gpu)

            # task 2 should have the same state
            self._compare_states(task.get_classy_state(), task_2.get_classy_state())

            # this tests that both states' iterators return the same samples
            sample = next(task.get_data_iterator())
            sample_2 = next(task_2.get_data_iterator())
            self._compare_samples(sample, sample_2)

            # test that the train step runs the same way on both states
            # and the loss remains the same
            task.train_step(use_gpu, local_variables)
            task_2.train_step(use_gpu, local_variables)
            self._compare_states(task.get_classy_state(), task_2.get_classy_state())

    def test_final_train_checkpoint(self):
        """Test that a train phase checkpoint with a where of 1.0 can be loaded"""

        config = get_fast_test_task_config()
        task = build_task(config).set_hooks(
            [CheckpointHook(self.base_dir, {}, phase_types=["train"])]
        )
        task_2 = build_task(config)

        use_gpu = torch.cuda.is_available()

        trainer = LocalTrainer(use_gpu=use_gpu)
        trainer.train(task)

        # load the final train checkpoint
        checkpoint = load_checkpoint(self.base_dir)

        # make sure fetching the where raises an exception, which means that
        # where is >= 1.0
        with self.assertRaises(Exception):
            task.where

        # set task_2's state as task's final train checkpoint
        task_2.set_checkpoint(checkpoint)
        task_2.prepare(use_gpu=use_gpu)

        # we should be able to train the task
        trainer.train(task_2)

    def test_test_only_checkpointing(self):
        """
        Tests checkpointing by running train_steps to make sure the
        train_steps run the same way after loading from a training
        task checkpoint on a test_only task.
        """
        train_config = get_fast_test_task_config()
        train_config["num_epochs"] = 10
        test_config = get_fast_test_task_config()
        test_config["test_only"] = True
        train_task = build_task(train_config).set_hooks([LossLrMeterLoggingHook()])
        test_only_task = build_task(test_config).set_hooks([LossLrMeterLoggingHook()])

        use_gpu = torch.cuda.is_available()

        # prepare the tasks for the right device
        train_task.prepare(use_gpu=use_gpu)

        # test in both train and test mode
        trainer = LocalTrainer(use_gpu=use_gpu)
        trainer.train(train_task)

        # set task's state as task_2's checkpoint
        test_only_task.set_checkpoint(
            get_checkpoint_dict(train_task, {}, deep_copy=True)
        )
        test_only_task.prepare(use_gpu=use_gpu)
        test_state = test_only_task.get_classy_state()

        # We expect the phase idx to be different for a test only task
        self.assertEqual(test_state["phase_idx"], -1)

        # We expect that test only state is test, no matter what train state is
        self.assertFalse(test_state["train"])

        # Num updates should be 0
        self.assertEqual(test_state["num_updates"], 0)

        # train_phase_idx should -1
        self.assertEqual(test_state["train_phase_idx"], -1)

        # Verify task will run
        trainer = LocalTrainer(use_gpu=use_gpu)
        trainer.train(test_only_task)

    @unittest.skipUnless(torch.cuda.is_available(), "This test needs a gpu to run")
    def test_checkpointing_different_device(self):
        config = get_fast_test_task_config()
        task = build_task(config)
        task_2 = build_task(config)

        for use_gpu in [True, False]:
            task.prepare(use_gpu=use_gpu)

            # set task's state as task_2's checkpoint
            task_2.set_checkpoint(get_checkpoint_dict(task, {}, deep_copy=True))

            # we should be able to run the trainer using state from a different device
            trainer = LocalTrainer(use_gpu=not use_gpu)
            trainer.train(task_2)



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import logging
import time
from typing import Any, Dict, Optional

from classy_vision import tasks
from classy_vision.generic.distributed_util import get_rank
from classy_vision.generic.perf_stats import PerfStats
from classy_vision.hooks.classy_hook import ClassyHook


class TimeMetricsHook(ClassyHook):
    """
    Computes and prints performance metrics. Logs at the end of a phase.
    """

    on_start = ClassyHook._noop
    on_forward = ClassyHook._noop
    on_step = ClassyHook._noop
    on_end = ClassyHook._noop

    def __init__(self, log_freq: Optional[int] = None) -> None:
        """
        Args:
            log_freq: if specified, logs every log_freq batches also.
        """
        super().__init__()
        self.log_freq: Optional[int] = log_freq
        self.start_time: Optional[float] = None

    def on_phase_start(
        self, task: "tasks.ClassyTask", local_variables: Dict[str, Any]
    ) -> None:
        """
        Initialize start time and reset perf stats
        """
        self.start_time = time.time()
        local_variables["perf_stats"] = PerfStats()

    def on_step(
        self, task: "tasks.ClassyTask", local_variables: Dict[str, Any]
    ) -> None:
        """
        Log metrics every log_freq batches, if log_freq is not None.
        """
        if self.log_freq is None:
            return
        batches = len(task.losses)
        if batches and batches % self.log_freq == 0:
            self._log_performance_metrics(task, local_variables)

    def on_phase_end(
        self, task: "tasks.ClassyTask", local_variables: Dict[str, Any]
    ) -> None:
        """
        Log metrics at the end of a phase if log_freq is None.
        """
        batches = len(task.losses)
        if batches:
            self._log_performance_metrics(task, local_variables)

    def _log_performance_metrics(
        self, task: "tasks.ClassyTask", local_variables: Dict[str, Any]
    ) -> None:
        """
        Compute and log performance metrics.
        """
        phase_type = task.phase_type
        batches = len(task.losses)

        if self.start_time is None:
            logging.warning("start_time not initialized")
        else:
            # Average batch time calculation
            total_batch_time = time.time() - self.start_time
            average_batch_time = total_batch_time / batches
            logging.info(
                "Average %s batch time (ms) for %d batches: %d"
                % (phase_type, batches, 1000.0 * average_batch_time)
            )

        # Train step time breakdown
        if local_variables.get("perf_stats") is None:
            logging.warning('"perf_stats" not set in local_variables')
        elif task.train:
            logging.info(
                "Train step time breakdown (rank {}):\n{}".format(
                    get_rank(), local_variables["perf_stats"].report_str()
                )
            )



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.


import sys


def debug_info(type, value, tb):
    if hasattr(sys, "ps1") or not sys.stderr.isatty():
        sys.__excepthook__(type, value, tb)
    else:
        import traceback
        import pdb

        traceback.print_exception(type, value, tb)
        print
        pdb.post_mortem(tb)



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import logging
from typing import Any, Dict

from classy_vision import tasks
from classy_vision.generic.distributed_util import is_master
from classy_vision.generic.visualize import plot_model
from classy_vision.hooks.classy_hook import ClassyHook


try:
    from tensorboardX import SummaryWriter  # noqa F401

    tbx_available = True
except ImportError:
    tbx_available = False


class ModelTensorboardHook(ClassyHook):
    """
    Shows the model graph in `TensorBoard <https
    ://www.tensorflow.org/tensorboard`>_.
    """

    on_phase_start = ClassyHook._noop
    on_forward = ClassyHook._noop
    on_loss_and_meter = ClassyHook._noop
    on_step = ClassyHook._noop
    on_phase_end = ClassyHook._noop
    on_end = ClassyHook._noop

    def __init__(self, tb_writer) -> None:
        """The constructor method of ModelTensorboardHook.

        Args:
            tb_writer: `Tensorboard SummaryWriter <https://tensorboardx.
            readthedocs.io/en/latest/tensorboard.html#tensorboardX.
            SummaryWriter>`_ instance

        """
        super().__init__()
        if not tbx_available:
            raise RuntimeError(
                "tensorboardX not installed, cannot use ModelTensorboardHook"
            )

        self.tb_writer = tb_writer

    def on_start(
        self, task: "tasks.ClassyTask", local_variables: Dict[str, Any]
    ) -> None:
        """
        Plot the model on Tensorboard.
        """
        if is_master():
            try:
                # Show model in tensorboard:
                logging.info("Showing model graph in TensorBoard...")

                plot_model(
                    task.base_model,
                    size=task.base_model.input_shape,
                    input_key=task.base_model.input_key
                    if hasattr(task.base_model, "input_key")
                    else None,
                    writer=self.tb_writer,
                )
            except Exception:
                logging.warn(
                    "Unable to plot model to tensorboard. Exception: ", exc_info=True
                )



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import unittest

from classy_vision.hooks import ClassyHook


class TestHook(ClassyHook):
    on_rendezvous = ClassyHook._noop
    on_start = ClassyHook._noop
    on_phase_start = ClassyHook._noop
    on_sample = ClassyHook._noop
    on_forward = ClassyHook._noop
    on_loss_and_meter = ClassyHook._noop
    on_backward = ClassyHook._noop
    on_step = ClassyHook._noop
    on_phase_end = ClassyHook._noop
    on_end = ClassyHook._noop

    def __init__(self, a, b):
        super().__init__()
        self.state.a = a
        self.state.b = b


class TestClassyHook(unittest.TestCase):
    def test_state_dict(self):
        a = 0
        b = {1: 2, 3: [4]}
        test_hook = TestHook(a, b)
        state_dict = test_hook.get_classy_state()
        # create a new test_hook and set its state to the old hook's.
        test_hook = TestHook("", 0)
        test_hook.set_classy_state(state_dict)
        self.assertEqual(test_hook.state.a, a)
        self.assertEqual(test_hook.state.b, b)
